"""Handle boolean values."""

from collections.abc import Mapping
from typing import Any

from pydantic import IPvAnyAddress
from pydantic.fields import FieldInfo

from fastlife.adapters.xcomponent.pydantic_form.widget_factory.base import (
    BaseWidgetBuilder,
)
from fastlife.adapters.xcomponent.pydantic_form.widgets.ip_address import (
    IpAddressWidget,
)


class IpAddrBuilder(BaseWidgetBuilder[IPvAnyAddress]):
    """Builder for boolean."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for boolean."""
        return issubclass(typ, IPvAnyAddress)

    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],
        field: FieldInfo | None,
        value: IPvAnyAddress | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> IpAddressWidget:
        """Build the widget."""
        return IpAddressWidget(
            name=field_name,
            removable=removable,
            title=field.title or "" if field else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            token=self.factory.token,
            value=value or None,
            error=form_errors.get(field_name),
        )



"""Types that are serialized over HTTP and forms."""

from decimal import Decimal
from typing import Any
from uuid import UUID

from pydantic.networks import EmailStr

Builtins = str | int | str | float | Decimal | UUID | EmailStr
"""Builtins types."""


AnyLiteral = Any
"""
Something like Literal[...] or Literal[...] which does not exists
or Idon't know where it is hidden.
"""



"""Widget base class."""

import json
import secrets
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Self, TypeVar

from markupsafe import Markup
from pydantic import Field, model_validator

from fastlife.adapters.xcomponent.registry import PYDANTICFORM_CATALOG_NS
from fastlife.domain.model.template import XTemplate

if TYPE_CHECKING:
    from fastlife.service.templates import AbstractTemplateRenderer

T = TypeVar("T")


class Widget(XTemplate, Generic[T]):
    """
    Base class for widget of pydantic fields.

    :param name: field name.
    :param value: field value.
    :param title: title for the widget.
    :param hint: hint for human.
    :param aria_label: html input aria-label value.
    :param value: current value.
    :param error: error of the value if any.
    :param children_types: childrens types list.
    :param removable: display a button to remove the widget for optional fields.
    :param token: token used to get unique id on the form.
    """

    namespace: ClassVar[str] = PYDANTICFORM_CATALOG_NS

    name: str
    "variable name, nested variables have dots."
    id: str = Field(default="")
    "variable name, nested variables have dots."
    value: T | None = Field(default=None)
    """Value of the field."""
    title: str = Field(default="")
    "Human title for the widget."
    hint: str | None = Field(default=None)
    "A help message for the the widget."

    error: str | None = Field(default=None)
    "Error message."

    aria_label: str | None = Field(default=None)
    "Non visible text alternative."
    token: str = Field(default="")
    "Unique token to ensure id are unique in the DOM."
    removable: bool = Field(default=False)
    "Indicate that the widget is removable from the dom."

    @model_validator(mode="after")
    def fill_props(self) -> Self:
        self.title = self.title or self.name.split(".")[-1]
        self.token = self.token or secrets.token_urlsafe(4).replace("_", "-")
        self.id = self.id or f"{self.name}-{self.token}".replace("_", "-").replace(
            ".", "-"
        )
        return self

    def to_html(self, renderer: "AbstractTemplateRenderer[XTemplate]") -> Markup:
        """Return the html version."""
        return Markup(renderer.render_template(self))


TWidget = TypeVar("TWidget", bound=Widget[Any])


class CustomWidget(Generic[TWidget]):
    """Type to annotate a field in order to override its default assitated widget."""

    typ: type[Any]

    def __init__(self, typ: type[TWidget], **kwargs: Any) -> None:
        self.typ = typ
        self.kwargs = kwargs or {}


class TypeWrapper:
    """
    Wrap children types for union type.

    :param typ: Wrapped type.
    :param route_prefix: route prefix used for ajax query to build type.
    :param name: name of the field wrapped.
    :param token: unique token to render unique id.
    :param title: title to display.

    """

    def __init__(
        self,
        typ: type[Any],
        route_prefix: str,
        name: str,
        token: str,
        title: str | None = None,
    ):
        self.typ = typ
        self.route_prefix = route_prefix
        self.name = name
        self.title = title or typ.__name__
        self.token = token

    @property
    def fullname(self) -> str:
        """Full name for the type."""
        return f"{self.typ.__module__}:{self.typ.__name__}"

    @property
    def id(self) -> str:
        """Unique id to inject in the DOM."""
        name = self.name.replace("_", "-").replace(".", "-").replace(":", "-")
        typ = self.typ.__name__.replace("_", "-")
        return f"{name}-{typ}-{self.token}"

    @property
    def params(self) -> str:
        """Params for the widget to render."""
        params = {
            "name": self.name,
            "token": self.token,
            "title": self.title,
        }
        return json.dumps(params)

    @property
    def url(self) -> str:
        """Url to fetch the widget."""
        ret = f"{self.route_prefix}/pydantic-form/widgets/{self.fullname}"
        return ret



"""Serialize session."""

import json
from base64 import b64decode, b64encode
from collections.abc import Mapping
from typing import Any

import itsdangerous

from fastlife.middlewares.session.serializer import AbsractSessionSerializer


class SignedSessionSerializer(AbsractSessionSerializer):
    """
    The default fastlife session serializer.

    It's based on the itsdangerous package to sign the session with a secret key.

    :param secret_key: a secret used to sign the session payload.

    :param max_age: session lifetime in seconds.
    """

    def __init__(self, secret_key: str, max_age: int) -> None:
        self.signer = itsdangerous.TimestampSigner(secret_key)
        self.max_age = max_age

    def serialize(self, data: Mapping[str, Any]) -> bytes:
        """Serialize and sign the session."""
        dump = json.dumps(data).encode("utf-8")
        encoded = b64encode(dump)
        signed = self.signer.sign(encoded)
        return signed

    def deserialize(self, data: bytes) -> tuple[Mapping[str, Any], bool]:
        """Deserialize the session.

        If the signature is incorect, the session restart from the begining.
        No exception raised.
        """
        try:
            data = self.signer.unsign(data, max_age=self.max_age)
            # We can't deserialize something wrong since the serialize
            # is signing the content.
            # If the signature key is compromise and we have invalid payload,
            # raising exceptions here is fine, it's dangerous afterall.
            session = json.loads(b64decode(data))
        except itsdangerous.BadSignature:
            return {}, True
        return session, False



"""Handle Literal type."""

from collections.abc import Mapping
from typing import Any, Literal

from pydantic.fields import FieldInfo

from fastlife.adapters.xcomponent.pydantic_form.widget_factory.base import (
    BaseWidgetBuilder,
)
from fastlife.adapters.xcomponent.pydantic_form.widgets.dropdown import DropDownWidget
from fastlife.adapters.xcomponent.pydantic_form.widgets.hidden import HiddenWidget
from fastlife.domain.model.types import AnyLiteral


class LiteralBuilder(BaseWidgetBuilder[AnyLiteral]):
    """Builder for Literal."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for Literal."""
        return origin is Literal

    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],  # a literal actually
        field: FieldInfo | None,
        value: AnyLiteral | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> HiddenWidget | DropDownWidget:
        """Build the widget."""
        choices: list[str] = field_type.__args__  # type: ignore
        if len(choices) == 1:
            return HiddenWidget(
                name=field_name,
                value=choices[0],
                token=self.factory.token,
            )
        return DropDownWidget(
            name=field_name,
            options=choices,  # type: ignore
            removable=removable,
            title=field.title or "" if field else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            token=self.factory.token,
            value=str(value),
            error=form_errors.get(field_name),
        )



"""Handle boolean values."""

from collections.abc import Mapping
from datetime import date, datetime
from typing import Any

from pydantic.fields import FieldInfo

from fastlife.adapters.xcomponent.pydantic_form.widget_factory.base import (
    BaseWidgetBuilder,
)
from fastlife.adapters.xcomponent.pydantic_form.widgets.date import (
    DateTimeWidget,
    DateWidget,
)


class DateBuilder(BaseWidgetBuilder[date]):
    """Builder for date."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for date."""
        return issubclass(typ, date)

    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],
        field: FieldInfo | None,
        value: date | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> DateWidget:
        """Build the widget."""
        return DateWidget(
            name=field_name,
            removable=removable,
            title=field.title or "" if field else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            token=self.factory.token,
            value=value,
            error=form_errors.get(field_name),
            # FIXME convert annotation to have min and max here
        )


class DateTimeBuilder(BaseWidgetBuilder[datetime]):
    """Builder for datetime."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for date."""
        return issubclass(typ, datetime)

    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],
        field: FieldInfo | None,
        value: datetime | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> DateTimeWidget:
        """Build the widget."""
        return DateTimeWidget(
            name=field_name,
            removable=removable,
            title=field.title or "" if field else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            token=self.factory.token,
            value=value,
            error=form_errors.get(field_name),
            # FIXME convert annotation to have min and max here
        )



"""Titles"""

from collections.abc import Mapping

from xcomponent import XNode

from fastlife.adapters.xcomponent.registry import BUILTINS_CATALOG_NS, x_component


@x_component(namespace=BUILTINS_CATALOG_NS)
def H1(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    <h1> html tag rendering

    :param children: child node.
    :param globals: injected root component globals variables.
    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.H1_CLASS`.
    """
    return """
        <h1 id={id} class={class_ or globals.H1_CLASS}>
            {children}
        </h1>
    """


@x_component(namespace=BUILTINS_CATALOG_NS)
def H2(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    <h2> html tag rendering

    :param children: child node.
    :param globals: injected root component globals variables.
    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.H2_CLASS`.
    """
    return """
        <h2 id={id} class={class_ or globals.H2_CLASS}>
            {children}
        </h2>
    """


@x_component(namespace=BUILTINS_CATALOG_NS)
def H3(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    <h3> html tag rendering

    :param children: child node.
    :param globals: injected root component globals variables.
    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.H3_CLASS`.
    """
    return """
        <h3 id={id} class={class_ or globals.H3_CLASS}>
            {children}
        </h3>
    """


@x_component(namespace=BUILTINS_CATALOG_NS)
def H4(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    <h4> html tag rendering

    :param children: child node.
    :param globals: injected root component globals variables.
    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.H4_CLASS`.
    """
    return """
        <h4 id={id} class={class_ or globals.H4_CLASS}>
            {children}
        </h4>
    """


@x_component(namespace=BUILTINS_CATALOG_NS)
def H5(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    <h5> html tag rendering

    :param children: child node.
    :param globals: injected root component globals variables.
    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.H5_CLASS`.
    """
    return """
        <h5 id={id} class={class_ or globals.H5_CLASS}>
            {children}
        </h5>
    """


@x_component(namespace=BUILTINS_CATALOG_NS)
def H6(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    <h6> html tag rendering

    :param children: child node.
    :param globals: injected root component globals variables.
    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.H6_CLASS`.
    """
    return """
        <h6 id={id} class={class_ or globals.H6_CLASS}>
            {children}
        </h6>
    """



"""
Service for scheduled jobs.
"""

from collections.abc import Awaitable, Callable
from datetime import datetime
from typing import Any, Generic, Literal, TypeAlias

from apscheduler.schedulers.base import BaseScheduler
from apscheduler.triggers.base import BaseTrigger
from apscheduler.util import _Undefined as Undefined  # type: ignore
from apscheduler.util import undefined

from fastlife.service.registry import TRegistry
from fastlife.shared_utils.resolver import resolve

__all__ = [
    "JobHandler",
    "JobScheduler",
    "JobSchedulerTrigger",
    "Undefined",
    "undefined",
]

JobSchedulerTriggerLiteral = Literal["interval", "cron", "date"]
JobSchedulerTrigger: TypeAlias = JobSchedulerTriggerLiteral | BaseTrigger


JobHandler = Callable[[TRegistry], None] | Callable[[TRegistry], Awaitable[None]]


class JobScheduler(Generic[TRegistry]):
    """
    Job scheduler and registry.
    """

    def __init__(self, registry: TRegistry) -> None:
        self.registry = registry
        self.scheduler: BaseScheduler = resolve(registry.settings.scheduler_class)()

    def register_job(
        self,
        job: JobHandler[TRegistry],
        /,
        *,
        trigger: JobSchedulerTrigger | None = None,
        id: str | None = None,
        name: str | None = None,
        misfire_grace_time: int | Undefined = undefined,
        coalesce: Undefined = undefined,
        max_instances: int | Undefined = undefined,
        next_run_time: datetime | Undefined = undefined,
        jobstore: str = "default",
        executor: str = "default",
        replace_existing: bool = False,
        **trigger_args: Any,
    ) -> None:
        """
        Register a scheduled job.

        :param job: the job handler to run.
        :param trigger: the way the job is triggered.
        :param id: optional identifier for the job.
        :param name: optional name for the job.
        :param misfire_grace_time: seconds after the designated runtime that the job is still
            allowed to be run (or `None` to allow the job to run no matter how late it is)
        :param coalesce: run once instead of many times if the scheduler determines that the
            job should be run more than once in succession
        :param max_instances: maximum number of concurrently running instances allowed for this
            job
        :param next_run_time: when to first run the job, regardless of the trigger (pass
            `None` to add the job as paused)
        :param jobstore: alias of the job store to store the job in
        :param executor: alias of the executor to run the job with
        :param replace_existing: `True` to replace an existing job with the same `id`
            (but retain the number of runs from the existing one)
        """
        self.scheduler.add_job(  # type: ignore
            job,
            kwargs={"registry": self.registry},
            trigger=trigger,
            id=id,
            name=name,
            misfire_grace_time=misfire_grace_time,
            coalesce=coalesce,
            max_instances=max_instances,
            next_run_time=next_run_time,
            jobstore=jobstore,
            executor=executor,
            replace_existing=replace_existing,
            **trigger_args,
        )

    def start(self) -> None:
        """Start the scheduler."""
        self.scheduler.start()

    def shutdown(self, *, wait: bool) -> None:
        """Shutdown the scheduler."""
        self.scheduler.shutdown(wait=wait)



"""
Prevents CSRF attack using cookie and html hidden field comparison.

Fast life did not reinvent the wheel on CSRF Protection.

It use the good old method. A CSRF token is saved in a cookie.
Forms post the CSRF token, and the token in the cookies and the form must match
to process the request, otherwise an exception
{class}`fastlife.service.csrf.CSRFAttack` is raised.

The cookie named is configurabllefia the settings
:attr:`fastlife.config.settings.Settings.csrf_token_name`

While using the `<Form/>` from XComponent or JinjaX, the csrf token is always sent.

The cookie is always set when you render any template. At the moment, there is
no way to prevent to set the cookie in the request.

"""

from collections.abc import Callable, Coroutine
from typing import Any

from fastlife.adapters.fastapi.request import Request


class CSRFAttack(Exception):
    """
    An exception raised if the cookie and the csrf token hidden field did not match.
    """


def check_csrf() -> Callable[[Request], Coroutine[Any, Any, bool]]:
    """
    A global application dependency, that is always active.

    If you don't want csrf token, its simple: don't use the
    application/x-www-form-urlencoded on a POST method.

    For security reason, there it no other options to disable this policy.

    :raises: {class}`.CSRFAttack` if the cookie and the csrf
        posted in the form does not match.
    """

    async def check_csrf(request: Request) -> bool:
        if (
            request.method != "POST"
            or request.headers.get("content-type")
            != "application/x-www-form-urlencoded"
        ):
            return True

        token = request.csrf_token
        form_data = await request.form()
        value = form_data.get(token.name)
        if value != token.value:
            raise CSRFAttack("CSRF token did not match")

        return True

    return check_csrf



"""Implement i18n."""

import pathlib
from collections import defaultdict
from collections.abc import Callable, Iterator, Mapping
from gettext import GNUTranslations
from io import BufferedReader
from typing import Any

from fastlife.shared_utils.resolver import resolve_path

LocaleName = str
Domain = str
CONTEXT_ENCODING = "%s\x04%s"


class TranslatableString(str):
    """
    Create a string made for translation associated to a domain.
    This class is instanciated by the
    {class}`fastlife.service.translations.TranslatableStringFactory` class.
    """

    __slots__ = ("domain",)

    def __new__(cls, msgid: str, domain: str) -> "TranslatableString":
        self = str.__new__(cls, msgid)
        self.domain = domain  # type: ignore
        return self


class TranslatableStringFactory:
    """Create a catalog of string associated to a domain."""

    def __init__(self, domain: str):
        self.domain = domain

    def __call__(self, msgid: str) -> str:
        """
        Use to generate the translatable string.

        usually:

        ```python
        _ = TranslatableStringFactory("mydomain")
        mymessage = _("translatable")
        ```

        Note that the string is associated to mydomain, so the babel extraction has
        to be initialized with that particular domain.
        """
        return TranslatableString(msgid, self.domain)


def find_mo_files(root_path: str) -> Iterator[tuple[LocaleName, Domain, pathlib.Path]]:
    """
    Find .mo files in a locales directory.

    :param root_path: locales directory.
    :return: a tupple containing locale_name, domain, file.
    """
    root = pathlib.Path(root_path)

    for locale_dir in root.iterdir():
        lc_messages_dir = locale_dir / "LC_MESSAGES"
        if not (locale_dir.is_dir() and lc_messages_dir.is_dir()):
            continue

        for mo_file in lc_messages_dir.glob("*.mo"):
            yield locale_dir.name, mo_file.stem, mo_file


def _default_plural(n: int) -> int:
    return int(n != 1)  # germanic plural by default


class MergedTranslations(GNUTranslations):
    _catalog: dict[str, str]

    def __init__(self) -> None:
        super().__init__()
        self._catalog = {}
        self.plural: Callable[[int], int] = _default_plural

    def merge(self, other: GNUTranslations) -> None:
        if hasattr(other, "_catalog"):
            self._catalog.update(other._catalog)  # type: ignore
        if hasattr(other, "plural"):
            self.plural = other.plural  # type: ignore


class Localizer:
    def __init__(self) -> None:
        self.translations: dict[Domain, MergedTranslations] = defaultdict(
            MergedTranslations
        )
        self.global_translations = MergedTranslations()

    def register(self, domain: str, file: BufferedReader) -> None:
        trans = GNUTranslations(file)
        self.translations[domain].merge(trans)
        self.global_translations.merge(trans)

    def as_dict(self) -> Mapping[str, Callable[..., str]]:
        return {
            "_": self.gettext,
            "gettext": self.gettext,
            "ngettext": self.ngettext,
            "dgettext": self.dgettext,
            "dngettext": self.dngettext,
            "pgettext": self.pgettext,
            "dpgettext": self.dpgettext,
            "npgettext": self.npgettext,
            "dnpgettext": self.dnpgettext,
        }

    def __call__(self, message: str, /, **mapping: Any) -> str:
        return self.gettext(message, **mapping)

    def gettext(self, message: str, /, **mapping: Any) -> str:
        if isinstance(message, TranslatableString):
            ret = self.translations[message.domain].gettext(message)  # type: ignore
        else:
            ret = self.global_translations.gettext(message)
        if mapping:
            ret = ret.format(**mapping)
        return ret

    def ngettext(self, singular: str, plural: str, n: int, /, **mapping: Any) -> str:
        ret = self.global_translations.ngettext(singular, plural, n)
        mapping_num = {"num": n, **mapping}
        return ret.format(**mapping_num)

    def dgettext(self, domain: str, message: str, /, **mapping: Any) -> str:
        ret = self.translations[domain].gettext(message)
        if mapping:
            ret = ret.format(**mapping)
        return ret

    def dngettext(
        self,
        domain: str,
        singular: str,
        plural: str,
        n: int,
        /,
        **mapping: Any,
    ) -> str:
        ret = self.translations[domain].ngettext(singular, plural, n)
        mapping_num = {"num": n, **mapping}
        return ret.format(**mapping_num)

    def pgettext(self, context: str, message: str, /, **mapping: Any) -> str:
        ret = self.global_translations.pgettext(context, message)
        if mapping:
            ret = ret.format(**mapping)
        return ret

    def dpgettext(
        self,
        domain: str,
        context: str,
        message: str,
        /,
        **mapping: Any,
    ) -> str:
        ret = self.translations[domain].pgettext(context, message)
        if mapping:
            ret = ret.format(**mapping)
        return ret

    def npgettext(
        self,
        context: str,
        singular: str,
        plural: str,
        n: int,
        /,
        **mapping: Any,
    ) -> str:
        ret = self.global_translations.npgettext(context, singular, plural, n)
        mapping_num = {"num": n, **mapping}
        return ret.format(**mapping_num)

    def dnpgettext(
        self,
        domain: str,
        context: str,
        singular: str,
        plural: str,
        n: int,
        /,
        **mapping: Any,
    ) -> str:
        ret = self.translations[domain].npgettext(context, singular, plural, n)
        mapping_num = {"num": n, **mapping}
        return ret.format(**mapping_num)


class TranslationDictionary:
    def __init__(self) -> None:
        self.translations: dict[LocaleName, Localizer] = defaultdict(Localizer)

    def load(self, root_path: str) -> None:
        for locale_name, domain, file_ in find_mo_files(root_path):
            with file_.open("rb") as stream:
                self.translations[locale_name].register(domain, stream)

    def get(self, locale_name: LocaleName) -> Localizer:
        return self.translations[locale_name]

    def __contains__(self, other: LocaleName) -> bool:
        return other in self.translations


class LocalizerFactory:
    """Initialize the proper translation context per request."""

    def __init__(self) -> None:
        self._translations = TranslationDictionary()
        self.null_localizer = Localizer()

    def load(self, path: str) -> None:
        """
        load a translations from the model.
        :param path: a python module and the locales dir separated by a `:`
        """
        root_path = resolve_path(path)
        self._translations.load(root_path)

    def __call__(self, locale_name: LocaleName) -> Localizer:
        """Create the translation context for the given request."""
        if locale_name not in self._translations:
            return self.null_localizer
        return self._translations.get(locale_name)



from pydantic import Field, IPvAnyAddress

from .base import Widget


class IpAddressWidget(Widget[IPvAnyAddress]):
    """
    Widget for Ip Address.
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div class="pt-4">
        <Label for={id}>{ globals.gettext(title) }</Label>
        <OptionalErrorText text={error} />
        <Input name={name} id={id}
            value={str(value)}
            type="text"
            pattern="^((\\d{1,3}\\.){3}\\d{1,3})|([0-9a-fA-F:]+)$"
            autocomplete={autocomplete}
            aria-label={aria_label} placeholder={placeholder} />
        <Hint text={hint} />
      </div>
    </Widget>
    """

    placeholder: str | None = Field(default=None)
    autocomplete: str | None = Field(default=None)
    """autocomplete attribute for the Input component."""



"""
This module load the html components for the xcomponent template engine.

Those components are loaded, injected in the catalog using a
{meth}`config.include <fastlife.config.configurator.Configurator.include>`
done before the app is created.
"""



"""HTTP Route."""

from collections.abc import Callable, Coroutine
from typing import Any

from fastapi.routing import APIRoute

from fastlife.domain.model.asgi import ASGIRequest, ASGIResponse


class Route(APIRoute):
    """
    Routing for fastlife application.

    The fastlife router construct fastlife request object in order to
    have the registry property available in every received request.
    """

    def get_route_handler(
        self,
    ) -> Callable[[ASGIRequest], Coroutine[Any, Any, ASGIResponse]]:
        """
        Replace the request object by the fastlife request associated with the registry.
        """
        orig_route_handler = super().get_route_handler()

        async def route_handler(request: ASGIRequest) -> ASGIResponse:
            req = request.scope["fastlife.registry"].request_factory(request)
            return await orig_route_handler(req)

        return route_handler



"""Security policy."""

import abc
from typing import Annotated, Any, Generic

from fastapi import Depends

from fastlife import GenericRequest, get_request
from fastlife.domain.model.security_policy import (
    Allowed,
    Anonymous,
    Authenticated,
    AuthenticationState,
    Forbidden,
    HasPermission,
    MFARequired,
    PendingMFA,
    TClaimedIdentity,
    TIdentity,
    Unauthorized,
)
from fastlife.service.registry import TRegistry


class AbstractSecurityPolicy(abc.ABC, Generic[TRegistry, TIdentity, TClaimedIdentity]):
    """Security policy base class."""

    Forbidden = Forbidden
    """The exception raised if the user identified is not granted."""
    Unauthorized = Unauthorized
    """The exception raised if no user has been identified."""
    MFARequired = MFARequired
    """The exception raised if no user has been authenticated using a MFA."""

    request: GenericRequest[TRegistry, TIdentity, TClaimedIdentity]
    """Request where the security policy is applied."""

    def __init__(
        self,
        request: Annotated[
            GenericRequest[TRegistry, TIdentity, TClaimedIdentity], Depends(get_request)
        ],
    ):
        """
        Build the security policy.

        When implementing a security policy, multiple parameters can be added
        to the constructor as FastAPI dependencies, using the `Depends` FastAPI
        annotation.
        The security policy is installed has a depenency of the router that hold
        a route prefix of the application.
        """
        self.request = request
        self.request.security_policy = self  # we do backref to implement has_permission
        self._authentication_state: (
            AuthenticationState[TClaimedIdentity, TIdentity] | None
        ) = None

    async def get_authentication_state(
        self,
    ) -> AuthenticationState[TClaimedIdentity, TIdentity]:
        """
        Return app-specific user object or None.
        """
        if self._authentication_state is None:
            self._authentication_state = await self.build_authentication_state()
        return self._authentication_state

    async def claimed_identity(self) -> TClaimedIdentity | None:
        """
        Return app-specific user object that pretend to be identified.
        """
        auth = await self.get_authentication_state()
        match auth:
            case PendingMFA(claimed):
                return claimed
            case _:
                return None

    async def identity(self) -> TIdentity | None:
        """
        Return app-specific user object after an mfa authentication or None.
        """
        auth = await self.get_authentication_state()
        match auth:
            case Authenticated(identity):
                return identity
            case _:
                return None

    @abc.abstractmethod
    async def build_authentication_state(
        self,
    ) -> AuthenticationState[TClaimedIdentity, TIdentity]:
        """
        Return the authentication state for the current request.
        """

    @abc.abstractmethod
    async def has_permission(
        self, permission: str
    ) -> HasPermission | type[HasPermission]:
        """Allow access to everything if signed in."""

    @abc.abstractmethod
    async def pre_remember(self, claimed_identity: TClaimedIdentity) -> None:
        """Save the user identity in the request session."""

    @abc.abstractmethod
    async def remember(self, identity: TIdentity) -> None:
        """Save the user identity in the request session."""

    @abc.abstractmethod
    async def forget(self) -> None:
        """Destroy the request session."""


class AbstractNoMFASecurityPolicy(AbstractSecurityPolicy[TRegistry, TIdentity, None]):
    async def pre_remember(self, claimed_identity: None) -> None:
        """Do Nothing."""


class InsecurePolicy(AbstractNoMFASecurityPolicy[Any, None]):
    """
    An implementation of the security policy made for explicit unsecured access.

    Setting a permission on a view require a security policy, if not set, accessing
    to a view will raise a RuntimeError. To bypass this error for testing purpose
    or your own reason, the InsecurePolicy has to be set to the configurator.
    """

    async def build_authentication_state(
        self,
    ) -> AuthenticationState[None, None]:
        return Anonymous

    async def has_permission(
        self, permission: str
    ) -> HasPermission | type[HasPermission]:
        """Access is allways granted."""
        return Allowed

    async def remember(self, identity: None) -> None:
        """Do nothing."""

    async def forget(self) -> None:
        """Do nothing."""



"""
API Resources declaration using a decorator.


```{literalinclude} ../../../../tests/fastlife_app/views/api/foos.py

```

"""

from collections.abc import Callable
from typing import Any

import venusian
from fastapi.types import IncEx

from fastlife.config.openapiextra import ExternalDocs

from .configurator import (
    VENUSIAN_CATEGORY,
    ConfigurationError,
    Configurator,
    OpenApiTag,
)


def resource(
    name: str,
    *,
    path: str | None = None,
    collection_path: str | None = None,
    description: str | None = None,
    external_docs: ExternalDocs | None = None,
) -> Callable[..., Any]:
    """
    A decorator to build an OpenAPI set of api grouped under a tag.

    This approach makes the resource the umbrella for all API methods that
    modify resources of that type, covering operations on both collections and
    individual items.

    The typical set of api route to manage resources is:

      * `POST /items`: a POST on a `collection_path` that create an item
      * `GET /items?filters=...&page=...`: a GET on a `collection_path` to list or
        even search items. Usually it returns partial item fields.
      * `GET /items/{id}`: Return the item with all its fields.
      * `PATCH /items/{id}`: An update of a single item on the resource `path`.
      * `DELETE /items/{id}`: Remove a single item from the collection on the `path`.

    You may implement more methods: `collection_get`, `collection_post`,
    `collection_put`, `collection_patch`, `collection_delete`, `collection_head`,
    `collection_options`, `get`, `post`, `put`, `patch`, `delete`, `head`, `options`
    will autmatically be bound to verb collection path or path.

    Note that there is no abstract class that declare this method, this is done by
    introspection while returning the configuration method
    {meth}`fastlife.config.configurator.GenericConfigurator.include`
    """
    tag = name

    def configure(
        wrapped: Callable[..., Any],
    ) -> Callable[..., Any]:
        def callback(scanner: venusian.Scanner, name: str, ob: type[Any]) -> None:
            if not hasattr(scanner, VENUSIAN_CATEGORY):
                return  # coverage: ignore

            config: Configurator = getattr(scanner, VENUSIAN_CATEGORY)
            if description:
                config.add_openapi_tag(
                    OpenApiTag(
                        name=tag, description=description, externalDocs=external_docs
                    )
                )

            api = ob()

            def bind_config(
                bind_config: Configurator,
                method: str,
                bind_path: str | None,
                endpoint: Any,
            ) -> None:
                if bind_path is None:
                    prefix = "collection_" if method.startswith("collection_") else ""
                    raise ConfigurationError(f"{prefix}path not set on resource {tag}")

                bind_config.add_api_route(
                    name=f"{method}_{tag}",
                    path=bind_path,
                    endpoint=endpoint,
                    tags=[tag],
                    methods=[method.split("_").pop()],
                    permission=endpoint.permission,
                    status_code=endpoint.status_code,
                    summary=endpoint.summary,
                    description=endpoint.description,
                    response_description=endpoint.response_description,
                    deprecated=endpoint.deprecated,
                    operation_id=endpoint.operation_id,
                    response_model=endpoint.response_model,
                    response_model_include=endpoint.response_model_include,
                    response_model_exclude=endpoint.response_model_exclude,
                    response_model_by_alias=endpoint.response_model_by_alias,
                    response_model_exclude_unset=endpoint.response_model_exclude_unset,
                    response_model_exclude_defaults=endpoint.response_model_exclude_defaults,
                    response_model_exclude_none=endpoint.response_model_exclude_none,
                    include_in_schema=endpoint.include_in_schema,
                    openapi_extra=endpoint.openapi_extra,
                )

            for method in dir(ob):
                match method:
                    case (
                        "collection_get"
                        | "collection_post"
                        | "collection_put"
                        | "collection_patch"
                        | "collection_delete"
                        | "collection_head"
                        | "collection_options"
                    ):
                        bind_config(
                            config, method, collection_path, getattr(api, method)
                        )
                    case (
                        "get" | "post" | "put" | "patch" | "delete" | "head" | "options"
                    ):
                        bind_config(config, method, path, getattr(api, method))
                    case _:
                        ...

        venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
        return wrapped

    return configure


def resource_view(
    *,
    permission: str | None = None,
    status_code: int | None = None,
    summary: str | None = None,
    description: str | None = None,
    response_description: str = "Successful Response",
    deprecated: bool | None = None,
    methods: list[str] | None = None,
    operation_id: str | None = None,
    response_model: type[Any] | None = None,
    response_model_include: IncEx | None = None,
    response_model_exclude: IncEx | None = None,
    response_model_by_alias: bool = True,
    response_model_exclude_unset: bool = False,
    response_model_exclude_defaults: bool = False,
    response_model_exclude_none: bool = False,
    include_in_schema: bool = True,
    openapi_extra: dict[str, Any] | None = None,
) -> Callable[..., Any]:
    """
    Decorator to use on a method of a class decorated with {func}`resource` in order
    to add OpenAPI information.

    This decorator has an effect if the decoratated method is named `collection_get`,
    `collection_post`, `collection_put`, `collection_patch`, `collection_delete`,
    `collection_head`, `collection_options`, `get`, `post`, `put`, `patch`, `delete`,
    `head` or `options`.

    :param permission: a permission to validate by the security policy.
    :param status_code: returned status_code
    :param summary: OpenAPI summary for the route.
    :param description:OpenAPI description for the route.
    :param response_description: OpenAPI description for the response.
    :param operation_id: OpenAPI optional unique string used to identify an
        operation.
    :param tags: OpenAPI tags for the route.
    :param deprecated: OpenAPI deprecated annotation for the route.
    :param openapi_extra: OpenAPI documentation extra fields.
    :param include_in_schema: Expose or not the route in the OpenAPI schema and
    documentation.

    :param response_model: class used for the api documentation for streaming response.
        It may also be used to validate data in case the view return dict.
        See [FastAPI doc](https://fastapi.tiangolo.com/tutorial/response-model/).
    :param response_model_include: customize fields list to include in response.
    :param response_model_exclude: customize fields list to exclude in response.
    :param response_model_by_alias: serialize fields by alias or by name if False.
    :param response_model_exclude_unset: exclude fields that are not explicitly
        set in response.
    :param response_model_exclude_defaults: exclude default value of response
        fields.
    :param response_model_exclude_none: exclude fields instead of serialize to
        null value.
    """

    def wrapped(fn: Any) -> Any:
        fn.permission = permission
        fn.status_code = status_code
        fn.summary = summary
        fn.description = description
        fn.response_description = response_description
        fn.deprecated = deprecated
        fn.methods = methods
        fn.operation_id = operation_id
        fn.response_model = response_model
        fn.response_model_include = response_model_include
        fn.response_model_exclude = response_model_exclude
        fn.response_model_by_alias = response_model_by_alias
        fn.response_model_exclude_unset = response_model_exclude_unset
        fn.response_model_exclude_defaults = response_model_exclude_defaults
        fn.response_model_exclude_none = response_model_exclude_none
        fn.include_in_schema = include_in_schema
        fn.openapi_extra = openapi_extra
        return fn

    return wrapped



from collections import defaultdict
from collections.abc import Callable

import venusian
from xcomponent import Catalog
from xcomponent.service.catalog import Component, Function

VENUSIAN_CATEGORY = "fastlife"  # copy/pasta from configurator
DEFAULT_CATALOG_NS = "app"
BUILTINS_CATALOG_NS = "builtins"
PYDANTICFORM_CATALOG_NS = "pydantic_form"

NSCatalog = dict[str, Catalog]

CatalogDict = dict[str, Component]
NSCatalogDict = dict[str, CatalogDict]


class XComponentRegistry:
    """
    Build the XComponent catalog using venusian.

    To avoid a global catalog, we generate the catalog by scanning using
    venusian.
    """

    components: NSCatalogDict

    functions: dict[str, Function]

    def __init__(self) -> None:
        self.components = defaultdict(dict)
        self.components[DEFAULT_CATALOG_NS] = {}  # ensure the app namespace exists
        self.functions = {}

    def register_xcomponent(
        self,
        name: str,
        component: Component,
        *,
        namespace: str,
    ) -> None:
        """Register a xcomponent component to the application template engine."""
        self.components[namespace][name] = component

    def register_xfunction(self, name: str, func: Function) -> None:
        """Register a xcomponent function to the application template engine."""
        self.functions[name] = func

    def build_catalogs(self) -> NSCatalog:
        catalogs: NSCatalog = {ns: Catalog() for ns in self.components.keys()}

        # we ensure we have a builtin catalog
        builtin_catalog: Catalog = catalogs.get(BUILTINS_CATALOG_NS) or Catalog()
        for name, component in self.components.get(BUILTINS_CATALOG_NS, {}).items():
            # we don't backref the builtins components to other namespace
            builtin_catalog.component(name, use=catalogs)(component)
        catalogs[BUILTINS_CATALOG_NS] = builtin_catalog

        for name, function in self.functions.items():
            builtin_catalog.function(name)(function)

        for ns, components in self.components.items():
            if ns == BUILTINS_CATALOG_NS:  # the default ns is process first and appart.
                continue

            ctlg = catalogs[ns]

            # we copy all the builtins components to all namespaces
            for name, component in self.components.get(BUILTINS_CATALOG_NS, {}).items():
                ctlg.component(name, use={})(component)

            for name, component in components.items():
                ctlg.component(name, use=catalogs)(component)

            for name, function in self.functions.items():
                ctlg.function(name)(function)

            catalogs[ns] = ctlg

        return catalogs


def x_component(
    namespace: str = DEFAULT_CATALOG_NS,
    name: str | None = None,
) -> Callable[[Component], Component]:
    """
    Register a component to the XComponent catalog for the application.

    It differ from from @catalog.component decorator of the xcomponent library,
    in the way the catalog is created.
    The fastlife configurator will register the component to a catalog during
    the configuration.

    :param name: override the name of the component, default is the python function
        name.
    :param use: import a catalog as a namespace components.
    """
    component_name = name

    def decorator(wrapped: Component) -> Component:
        def callback(
            scanner: venusian.Scanner,
            name: str,
            ob: Component,
        ) -> None:
            if not hasattr(scanner, VENUSIAN_CATEGORY):
                return  # coverage: ignore
            scanner.fastlife.register_xcomponent(  # type: ignore
                component_name or name,
                ob,
                namespace=namespace,
            )

        venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
        return wrapped

    return decorator


def x_function(name: str | None = None) -> Callable[[Function], Function]:
    """
    Register a component to the XComponent function for the application.

    It differ from from @catalog.function decorator of the xcomponent library,
    in the way the catalog is created.
    The fastlife configurator will register the function to a catalog during
    the configuration.

    :param name: override the name of the function, default is the python function name.
    """
    function_name = name

    def decorator(wrapped: Function) -> Function:
        def callback(
            scanner: venusian.Scanner,
            name: str,
            ob: Component,
        ) -> None:
            if not hasattr(scanner, VENUSIAN_CATEGORY):
                return  # coverage: ignore
            scanner.fastlife.register_xfunction(function_name or name, ob)  # type: ignore

        venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
        return wrapped

    return decorator



"""Session store for the webtest client."""

import time
from collections.abc import Mapping
from http.cookiejar import Cookie
from typing import TYPE_CHECKING, Any

import httpx

if TYPE_CHECKING:
    from .testclient import WebTestClient  # coverage: ignore


CookieTypes = httpx._types.CookieTypes  # type: ignore
Cookies = httpx._models.Cookies  # type: ignore


class Session(dict[str, Any]):
    """Manipulate the session of the WebTestClient browser."""

    def __init__(self, client: "WebTestClient"):
        self.client = client
        self.srlz = client.session_serializer
        self.settings = self.client.settings
        data: Mapping[str, Any]
        cookie_name = self.settings.session_cookie_name
        self.has_session = cookie_name in self.client.cookies
        if self.has_session:
            data, error = self.srlz.deserialize(
                self.client.cookies[cookie_name].encode("utf-8")
            )
            if error:
                self.has_session = False
        else:
            data = {}
        super().__init__(data)

    def __setitem__(self, __key: Any, __value: Any) -> None:
        """Initialize a value in the session of the client in order to test."""
        super().__setitem__(__key, __value)
        settings = self.settings
        data = self.serialize()
        self.client.cookies.jar.set_cookie(
            Cookie(
                version=0,
                name=settings.session_cookie_name,
                value=data,
                port=None,
                port_specified=False,
                domain=f".{settings.session_cookie_domain}",
                domain_specified=True,
                domain_initial_dot=True,
                path="/",
                path_specified=True,
                secure=False,
                expires=int(time.time() + settings.session_duration.total_seconds()),
                discard=False,
                comment=None,
                comment_url=None,
                rest={"HttpOnly": None, "SameSite": "lax"},  # type: ignore
                rfc2109=False,
            )
        )

    def serialize(self) -> str:
        """Serialize the session"""
        return self.srlz.serialize(self).decode("utf-8")



"""Find the localization gor the given request."""

from collections.abc import Callable
from typing import Any

from fastlife.settings import Settings

LocaleName = str
"""The LocaleName is a locale such as en, fr that will be consume for translations."""

from fastlife.adapters.fastapi.request import GenericRequest

LocaleNegociator = Callable[[GenericRequest[Any, Any, Any]], LocaleName]
"""Interface to implement to negociate a locale"""


def default_negociator(settings: Settings) -> LocaleNegociator:
    """The default local negociator return the locale set in the conf."""

    def locale_negociator(request: "GenericRequest[Any, Any, Any]") -> str:
        return settings.default_locale

    return locale_negociator



"""ASGI types from Starlette."""

from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp, Message, Receive, Scope, Send

ASGIRequest = Request
"""Starlette request class used as ASGI Protocol base HTTP Request representation."""

ASGIResponse = Response
"""Starlette request class used as ASGI Protocol base HTTP Response representation."""

__all__ = [
    "ASGIApp",
    "ASGIRequest",
    "ASGIResponse",
    "Message",
    "Receive",
    "Scope",
    "Send",
]



from collections.abc import AsyncIterator, Awaitable, Iterable
from typing import Annotated, Any, Generic

from fastapi import Depends
from fastapi.websockets import WebSocket as BaseWebSocket
from starlette.responses import Response

from fastlife.service.registry import DefaultRegistry, TRegistry


class MyWebSocket(Generic[TRegistry]):
    registry: TRegistry

    def __init__(self, socket: BaseWebSocket) -> None:
        self._sock = socket
        self.registry = socket.scope["fastlife.registry"]

    async def accept(
        self,
        subprotocol: str | None = None,
        headers: Iterable[tuple[bytes, bytes]] | None = None,
    ) -> None:
        return await self._sock.accept(subprotocol, headers)

    def receive_text(self) -> Awaitable[str]:
        return self._sock.receive_text()

    def receive_bytes(self) -> Awaitable[bytes]:
        return self._sock.receive_bytes()

    def receive_json(self, mode: str = "text") -> Awaitable[Any]:
        return self._sock.receive_json(mode)

    def iter_text(self) -> AsyncIterator[str]:
        return self._sock.iter_text()  # coverage: ignore

    def iter_bytes(self) -> AsyncIterator[bytes]:
        return self._sock.iter_bytes()  # coverage: ignore

    def iter_json(self) -> AsyncIterator[Any]:
        return self._sock.iter_json()  # coverage: ignore

    def send_text(self, data: str) -> Awaitable[None]:
        return self._sock.send_text(data)

    def send_bytes(self, data: bytes) -> Awaitable[None]:
        return self._sock.send_bytes(data)

    def send_json(self, data: Any, mode: str = "text") -> Awaitable[None]:
        return self._sock.send_json(data, mode)

    def close(self, code: int = 1000, reason: str | None = None) -> Awaitable[None]:
        return self._sock.close()

    def send_denial_response(self, response: Response) -> Awaitable[None]:
        return self._sock.send_denial_response(response)  # coverage: ignore


def get_websocket(socket: BaseWebSocket) -> MyWebSocket[Any]:
    return MyWebSocket(socket)


WebSocket = Annotated[MyWebSocket[TRegistry], Depends(get_websocket)]
GenericWebSocket = WebSocket[DefaultRegistry]

__all__ = ["WebSocket", "GenericWebSocket"]



"""
Widget for field of type Union.
"""

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING

from markupsafe import Markup
from pydantic import BaseModel, Field

from fastlife.domain.model.template import XTemplate

if TYPE_CHECKING:
    from fastlife.service.templates import AbstractTemplateRenderer

from .base import TWidget, TypeWrapper, Widget


class UnionWidget(Widget[TWidget]):
    """
    Widget for union types.
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div id={id}>
        <Details>
          <Summary id={id + '-union-summary'}>
            <H3 class={globals.H3_SUMMARY_CLASS}>{ globals.gettext(title) }</H3>
            <OptionalErrorText text={error} />
          </Summary>
          <div hx-sync="this" id={id + "-child"} class="flex flex-row gap-2 mt-2">
            {
              if child {
                child
              }
              else {
                for typ in types {
                  <Button type="button"
                    hx-target="closest div"
                    hx-get={typ.url}
                    hx-vals={typ.params}
                    id={typ.id}
                    onclick={"document.getElementById('" + id + "-remove-btn').hidden=false"}
                    class={globals.SECONDARY_BUTTON_CLASS}>{globals.gettext(typ.title)}</Button>
                }
              }
            }
          </div>
          <div class="p-4">
            <Button
                type="button"
                id={id + '-remove-btn'}
                hidden={not child}
                class={globals.SECONDARY_BUTTON_CLASS}
                onclick={'resetUnion("' + id + '")'}
                >
                {globals.gettext("Remove")}
            </Button>
            <script>
                function resetUnion(id) {
                    const child = document.getElementById(id + "-child");
                    const defaultBtns = document.getElementById(id + "-default-buttons");
                    const btn = document.getElementById(id + '-remove-btn');
                    child.innerHTML = defaultBtns.innerHTML;
                    htmx.process(child);
                    btn.hidden = true;
                }
            </script>
          </div>
          <div class="hidden" id={id + "-default-buttons"}>
            {
              for typ in types {
                <Button type="button"
                  hx-target="closest div"
                  hx-get={typ.url}
                  hx-vals={typ.params}
                  id={typ.id}
                  onclick={"document.getElementById('" + id + "-remove-btn').hidden=false"}
                  class={globals.SECONDARY_BUTTON_CLASS}>{globals.gettext(typ.title)}</Button>
              }
            }
          </div>
        </Details>
      </div>
    </Widget>
    """

    children_types: Mapping[str, type[BaseModel]]

    types: Sequence[TypeWrapper] | None = Field(default=None)
    child: str = Field(default="")

    def build_types(self, route_prefix: str) -> Sequence[TypeWrapper]:
        """Wrap types in the union in order to get the in their own widgets."""
        return [
            TypeWrapper(typ, route_prefix, self.name, self.token, title)
            for title, typ in self.children_types.items()
        ]

    def to_html(self, renderer: "AbstractTemplateRenderer[XTemplate]") -> Markup:
        """Return the html version."""
        self.child = Markup(self.value.to_html(renderer)) if self.value else ""
        self.types = self.build_types(renderer.route_prefix)
        return Markup(renderer.render_template(self))



"""Serialize session."""

import abc
from collections.abc import Mapping
from typing import Any


class AbsractSessionSerializer(abc.ABC):
    """Session serializer base class"""

    @abc.abstractmethod
    def __init__(self, secret_key: str, max_age: int) -> None: ...

    @abc.abstractmethod
    def serialize(self, data: Mapping[str, Any]) -> bytes:
        """Serialize the session content to bytes in order to be saved."""
        ...

    @abc.abstractmethod
    def deserialize(self, data: bytes) -> tuple[Mapping[str, Any], bool]:
        """Derialize the session raw bytes content and return it as a mapping."""
        ...



from collections.abc import Mapping
from urllib.parse import quote

from starlette.background import BackgroundTask
from starlette.datastructures import URL
from starlette.responses import Response


class RedirectResponse(Response):
    """
    A redirect response for Post/Redirect/Get pattern.

    The starlette default value status code is 307, which means that it is used
    as a way to replay the same query which is definitly not the most used case
    in web applications.

    This is why the redirect response here is using 303 see other which
    ensure a GET request will be made for the redirection.

    A new parameter hx_redirect exists in order to set the HX-Redirect header
    to follow a browser redirection from an ajax query.
    """

    def __init__(
        self,
        url: str | URL,
        hx_redirect: bool = False,
        status_code: int = 303,
        headers: Mapping[str, str] | None = None,
        background: BackgroundTask | None = None,
    ):
        super().__init__(
            content=b"", status_code=status_code, headers=headers, background=background
        )
        self.headers["HX-Redirect" if hx_redirect else "location"] = quote(
            str(url), safe=":/%#?=@[]!$&'()*+,;"
        )


__all__ = ["Response", "RedirectResponse"]



from typing import Generic

from fastlife import Configurator, configure
from fastlife.domain.model.asgi import ASGIApp, Receive, Scope, Send
from fastlife.middlewares.base import AbstractMiddleware
from fastlife.service.registry import GenericRegistry, TSettings


class RegistryMiddleware(AbstractMiddleware, Generic[TSettings]):
    def __init__(self, app: ASGIApp, *, registry: GenericRegistry[TSettings]) -> None:
        self.app = app
        self._registry: GenericRegistry[TSettings] = registry

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Called every time an http request is reveived.

        This method before the request object even exists.
        """
        scope["fastlife.registry"] = self._registry
        await self.app(scope, receive, send)


@configure
def includeme(config: Configurator) -> None:
    config.add_middleware(RegistryMiddleware, registry=config.registry)



"""
Template renderer based on XComponent.

More template engine can be registered using the configurator method
{meth}`add_renderer <fastlife.config.configurator.GenericConfigurator.add_renderer>`
"""

from collections.abc import Mapping
from typing import Any

from markupsafe import Markup
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from xcomponent import Catalog

from fastlife import Request, XTemplate
from fastlife.adapters.xcomponent.pydantic_form.widget_factory.factory import (
    WidgetFactory,
)
from fastlife.config import Configurator, configure
from fastlife.domain.model.form import FormModel
from fastlife.service.templates import (
    AbstractTemplateRenderer,
    AbstractTemplateRendererFactory,
)
from fastlife.settings import Settings
from fastlife.shared_utils.resolver import resolve


class XTemplateRenderer(AbstractTemplateRenderer[XTemplate]):
    """
    An object that will be initialized by an AbstractTemplateRendererFactory,
    passing the request to process.
    """

    request: Request
    """Associated request that needs a response."""
    ns_catalog: dict[str, Catalog]

    def __init__(
        self, globals: dict[str, Any], request: Request, ns_catalog: dict[str, Catalog]
    ) -> None:
        self.request = request
        self.ns_catalog = ns_catalog
        self.globals: dict[str, Any] = {**globals}
        self.globals["pydantic_form"] = self.pydantic_form

    def pydantic_form(
        self, model: FormModel[Any], *, token: str | None = None
    ) -> Markup:
        """
        Generate HTML markup to build a form from the given form model.

        :param model: the form model that will be transformed to markup.
        :param token: a token used to ensure that unique identifier are unique.
        :return: HTML Markup generated by composign fields widgets.
        """
        return WidgetFactory(self, token).get_markup(model)

    def pydantic_form_field(
        self,
        model: type[Any],
        *,
        name: str | None,
        token: str | None,
        removable: bool,
        field: FieldInfo | None,
    ) -> Markup:
        """
        Generate HTML for a particular field in a form.

        This function is used to generate union subtypes in Ajax requests.
        :param model: a pydantic or python builtin type that is requests to be rendered
        :param name: name for the field
        :param token: the token of the form to render unique identifiers
        :param removable: add a way let the user remove the widget
        :param field: only render this particular field for the model.
        :return: HTML Markup.
        """
        prefix = name or self.request.registry.settings.form_data_model_prefix
        form_data: Mapping[str, Any] = {}
        if issubclass(model, BaseModel):
            form_data = FormModel[Any].default(prefix, model).form_data
        return (
            WidgetFactory(self, token)
            .get_widget(
                model,
                form_data=form_data,
                form_errors={},
                prefix=prefix,
                removable=removable,
                field=field,
            )
            .to_html(self)
        )

    def render_template(self, template: XTemplate) -> str:
        """
        Render an inline template.

        :param template: the template to render.
        :return: The template rendering result.
        """
        params = {
            k: getattr(template, k) for k in template.__class__.model_fields.keys()
        }
        return self.ns_catalog[template.namespace].render(
            template.template,
            globals=self.globals,
            **params,
        )


class XRendererFactory(AbstractTemplateRendererFactory[XTemplate]):
    """
    The template render factory.
    """

    def __init__(self, settings: "Settings", ns_catalog: dict[str, Catalog]) -> None:
        self.globals = resolve(settings.xcomponent_global_catalog_class)().model_dump()
        self.ns_catalog = ns_catalog

    def __call__(self, request: Request) -> AbstractTemplateRenderer[XTemplate]:
        """
        While processing an HTTP Request, a renderer object is created giving
        isolated context per request.

        :param Request: the HTTP Request to process.
        :return: The renderer object that will process that request.
        """
        return XTemplateRenderer(
            globals=self.globals,
            request=request,
            ns_catalog=self.ns_catalog,
        )


@configure
def includeme(conf: Configurator) -> None:
    conf.include(".functions")
    conf.include(".html")
    conf.include(".icons")
    conf.include(".pydantic_form")
    conf.add_renderer(
        XRendererFactory(
            conf.registry.settings,
            conf.build_catalogs(),
        ),
    )



"""Pydantic models"""

from collections.abc import Sequence
from typing import TYPE_CHECKING

from markupsafe import Markup
from pydantic import Field

from fastlife.domain.model.template import XTemplate

if TYPE_CHECKING:
    from fastlife.service.templates import AbstractTemplateRenderer

from .base import TWidget, Widget


class ModelWidget(Widget[Sequence[TWidget]]):
    template = """
    <Widget widget_id={id} removable={removable}>
    <div id={id} class={if nested {"m-4"}}>
        {
          if nested {
            <Details>
              <Summary id={id + '-summary'}>
                <H3 class={globals.H3_SUMMARY_CLASS}>{ globals.gettext(title) }</H3>
                <OptionalErrorText text={error} />
              </Summary>
              <div>
                {
                  for child in children_widgets {
                    child
                  }
                }
              </div>
            </Details>
          }
          else {
            for child in children_widgets {
              child
            }
          }
        }
    </div>
    </Widget>
    """

    nested: bool = Field(default=False)
    children_widgets: list[str] | None = Field(default=None)

    def to_html(self, renderer: "AbstractTemplateRenderer[XTemplate]") -> Markup:
        """Return the html version."""
        self.children_widgets = [child.to_html(renderer) for child in self.value or []]
        return Markup(renderer.render_template(self))



"""
Base class to of the template renderer.

Fastlife comes with the one rendering engine
{class}`fastlife.adapters.xcomponent.renderer.XRendererFactory`.

More template engine can be registered using the configurator method
{meth}`add_renderer <fastlife.config.configurator.GenericConfigurator.add_renderer>`
"""

import abc
from collections.abc import Mapping
from typing import Any, Generic, TypeVar, get_args, get_origin

from markupsafe import Markup
from pydantic.fields import FieldInfo

from fastlife import Request, Response
from fastlife.domain.model.form import FormModel
from fastlife.domain.model.template import InlineTemplate

TInlineTemplate = TypeVar("TInlineTemplate", bound=InlineTemplate)


class AbstractTemplateRenderer(abc.ABC, Generic[TInlineTemplate]):
    """
    An object that will be initialized by an AbstractTemplateRendererFactory,
    passing the request to process.
    """

    request: Request
    """Associated request that needs a response."""

    def __init__(self, request: Request) -> None:
        self.request = request
        self.globals: dict[str, Any] = {}

    @property
    def route_prefix(self) -> str:
        """Used to buid pydantic form widget that do ajax requests."""
        return self.request.registry.settings.fastlife_route_prefix

    def render(
        self,
        template: str,
        *,
        status_code: int = 200,
        content_type: str = "text/html",
        globals: Mapping[str, Any] | None = None,
        params: TInlineTemplate,
    ) -> Response:
        """
        Render the template and build the HTTP Response.
        """
        request = self.request
        if globals:
            self.globals.update(globals)
        data = self.render_template(params)
        resp = Response(
            data, status_code=status_code, headers={"Content-Type": content_type}
        )
        resp.set_cookie(
            request.csrf_token.name,
            request.csrf_token.value,
            httponly=True,
            secure=request.url.scheme == "https",
            samesite="strict",
            max_age=60 * 15,
        )
        return resp

    @abc.abstractmethod
    def render_template(self, template: TInlineTemplate) -> str:
        """
        Render an inline template.

        :param template: the template to render.
        :return: The template rendering result.
        """

    @abc.abstractmethod
    def pydantic_form(
        self, model: FormModel[Any], *, token: str | None = None
    ) -> Markup:
        """
        Generate HTML markup to build a form from the given form model.

        :param model: the form model that will be transformed to markup.
        :param token: a token used to ensure that unique identifier are unique.
        :return: HTML Markup generated by composign fields widgets.
        """

    @abc.abstractmethod
    def pydantic_form_field(
        self,
        model: type[Any],
        *,
        name: str | None,
        token: str | None,
        removable: bool,
        field: FieldInfo | None,
    ) -> Markup:
        """
        Generate HTML for a particular field in a form.

        This function is used to generate union subtypes in Ajax requests.
        :param model: a pydantic or python builtin type that is requests to be rendered
        :param name: name for the field
        :param token: the token of the form to render unique identifiers
        :param removable: add a way let the user remove the widget
        :param field: only render this particular field for the model.
        :return: HTML Markup.
        """


class AbstractTemplateRendererFactory(abc.ABC, Generic[TInlineTemplate]):
    """
    The template render factory.
    """

    template_type: type[TInlineTemplate]

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for base in cls.__orig_bases__:  # type: ignore
            if get_origin(base) is AbstractTemplateRendererFactory:  # type: ignore
                (cls.template_type,) = get_args(base)

    @abc.abstractmethod
    def __call__(self, request: Request) -> AbstractTemplateRenderer[TInlineTemplate]:
        """
        While processing an HTTP Request, a renderer object is created giving
        isolated context per request.

        :param Request: the HTTP Request to process.
        :return: The renderer object that will process that request.
        """



"""
Widget for field of type bool.
"""

from .base import Widget


class BooleanWidget(Widget[bool]):
    """
    Widget for field of type bool.
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div class="pt-4">
        <div class="flex items-center">
         <Checkbox name={name} id={id} checked={value} value="1" />
         <Label for={id} class="ms-2 text-base text-neutral-900 dark:text-white">
            {globals.gettext(title)}
          </Label>
        </div>
       <OptionalErrorText text={error} />
      </div>
    </Widget>
    """



"""Domain model of the typer integration."""

from collections.abc import Callable
from typing import Any

from typer.core import TyperCommand
from typer.models import CommandInfo, Default

CLIHook = Callable[..., Any]


class CLICommand:
    """Keep CLI command arguments and function for the final build."""

    hook: CLIHook
    cmd: CommandInfo

    def __init__(
        self,
        hook: CLIHook,
        name: str | None = None,
        cls: type[TyperCommand] | None = None,
        context_settings: dict[Any, Any] | None = None,
        help: str | None = None,
        epilog: str | None = None,
        short_help: str | None = None,
        options_metavar: str | None = None,
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
        rich_help_panel: str | None = Default(None),
    ) -> None:
        self.hook = hook
        self.cmd = CommandInfo(
            name=name,
            cls=cls,
            context_settings=context_settings,
            help=help,
            epilog=epilog,
            short_help=short_help,
            options_metavar=options_metavar or "[OPTIONS]",
            add_help_option=add_help_option,
            no_args_is_help=no_args_is_help,
            hidden=hidden,
            deprecated=deprecated,
            rich_help_panel=rich_help_panel,
        )



from datetime import date, datetime
from typing import Any

from fastlife.adapters.xcomponent.registry import x_function


@x_function(name="len")
def length(sized: Any) -> int:
    return len(sized)


@x_function()
def is_bool(i: Any) -> bool:
    return isinstance(i, bool)


@x_function()
def is_str(value: Any) -> bool:
    return isinstance(value, str)


@x_function()
def isoformat(value: str | date | datetime | None) -> str | None:
    """
    Cast a date or a datetime to isoformat representation.

    In case its not a date or a datetime, the value is preserved.
    """
    if isinstance(value, datetime):
        return value.replace(tzinfo=None).isoformat()
    if isinstance(value, date):
        return value.isoformat()
    return value


@x_function(name="str")
def to_string(value: Any) -> str:
    """convert a value to a string, the None value is displayed as an empty string."""
    if value is None:
        return ""
    return str(value)



"""HTTP Form serialization."""

from collections.abc import Mapping
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, ValidationError

from fastlife.shared_utils.form import flatten_error

T = TypeVar("T", bound=BaseModel)
"""Template type for form serialized model"""


class FormModel(Generic[T]):
    prefix: str
    model: T
    fatal_error: str
    errors: dict[str, str]
    is_valid: bool

    def __init__(
        self, prefix: str, model: T, errors: dict[str, Any], is_valid: bool = False
    ) -> None:
        self.prefix = prefix
        self.model = model
        self.fatal_error = ""
        self.errors = errors
        self.is_valid = is_valid

    @classmethod
    def default(cls, prefix: str, pydantic_type: type[T]) -> "FormModel[T]":
        return cls(prefix, pydantic_type.model_construct(), {})

    def set_fatal_error(self, value: str) -> None:
        self.fatal_error = value
        self.is_valid = False

    def add_error(self, field: str, value: str) -> None:
        self.errors[f"{self.prefix}.{field}"] = value
        self.is_valid = False

    def edit(self, pydantic_type: T) -> None:
        """
        Load the form with the given model and consider it as valid for the user.

        No error will be reported.
        """
        self.model = pydantic_type
        self.errors = {}
        self.is_valid = True

    @property
    def form_data(self) -> Mapping[str, Any]:
        return {self.prefix: self.model.model_dump(warnings="none")}

    @classmethod
    def from_payload(
        cls, prefix: str, pydantic_type: type[T], data: Mapping[str, Any]
    ) -> "FormModel[T]":
        try:
            ptyp = pydantic_type(**data.get(prefix, {}))
            ret = cls(prefix, ptyp, {}, True)
            return ret
        except ValidationError as exc:
            errors: dict[str, str] = flatten_error(exc, prefix, pydantic_type)
            model = pydantic_type.model_construct(**data.get(prefix, {}))
            return cls(prefix, model, errors)



"""Collapsible components, using details/summary."""

from collections.abc import Mapping

from xcomponent import XNode

from fastlife.adapters.xcomponent.registry import BUILTINS_CATALOG_NS, x_component


@x_component(namespace=BUILTINS_CATALOG_NS)
def Details(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
    open: bool = True,
) -> str:
    """
    Produce a ``<details>`` html node in order to create a collapsible box.

    .. code-block:: html

      <Details>
        <Summary id="my-summary">
          <H3 class={globals.H3_SUMMARY_CLASS}>A title</H3>
        </Summary>
        <div>
          Some content
        </div>
      </Details>

    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.DETAILS_CLASS`.
    :param open: open/close state.
    """
    return """
    <details id={id} class={class_ or globals.DETAILS_CLASS} open={open}>
        {children}
    </details>
    """


@x_component(namespace=BUILTINS_CATALOG_NS)
def Summary(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
    open: bool = True,
) -> str:
    """
    Create html ``<summary>`` node for the
    :func:`fastlife.adapters.xcomponent.html.collapsible.Details` component.

    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.SUMMARY_CLASS`.
    :param open: Open or collapse the content of the details.
    """
    return """
    <summary id={id} class={class_ or globals.SUMMARY_CLASS}
      style="list-style: none; -webkit-user-select: none; -moz-user-select: none; -ms-user-select: none; user-select: none;"
      onclick={"document.getElementById('" + id + "-icon').classList.toggle('rotate-90')"}>
      <Icon name="chevron-right" id={id + '-icon'}
        class={
            if open {
                <>
                    {globals.SUMMARY_OPEN_ICON_CLASS}
                </>
            }
            else {
                <>
                    {globals.SUMMARY_CLOSED_ICON_CLASS}
                </>
            }
        } />
        { children }
    </summary>
    """



"""Security policy."""

import logging
from collections.abc import Callable, Coroutine
from typing import Any, Generic, Literal, TypeVar

from fastapi import HTTPException
from starlette.status import HTTP_401_UNAUTHORIZED, HTTP_403_FORBIDDEN

CheckPermissionHook = Callable[..., Coroutine[Any, Any, None]] | Callable[..., None]
CheckPermission = Callable[[str], CheckPermissionHook]

TClaimedIdentity = TypeVar("TClaimedIdentity")
TIdentity = TypeVar("TIdentity")

log = logging.getLogger(__name__)


class _Anonymous: ...


Anonymous = _Anonymous()
"""
The user is not authenticated.
"""


class PendingMFA(Generic[TClaimedIdentity]):
    """
    The user provided its identity, usually validated with a first factor,
    such as a password but it has not totally proved its authentication
    by a second or many other factors of authentication.
    The type TClaimedIdentity will store the relevant informations during
    this authentication phase.
    """

    claimed: TClaimedIdentity | None
    __match_args__ = ("claimed",)

    def __init__(self, claimed: TClaimedIdentity) -> None:
        self.claimed = claimed


class Authenticated(Generic[TIdentity]):
    """The identity has been validated."""

    __match_args__ = ("identity",)

    def __init__(self, identity: TIdentity) -> None:
        self.identity = identity


AuthenticationState = (
    _Anonymous | PendingMFA[TClaimedIdentity] | Authenticated[TIdentity]
)
"""
Type representing the state of an authentication.
"""

NoMFAAuthenticationState = AuthenticationState[None, TIdentity]
"""
Type representing a state of authentication when no multiple factor of authentication
is involved.
"""


class Unauthorized(HTTPException):
    """An exception raised to stop a request execution and return a 401 HTTP Error."""

    def __init__(
        self,
        status_code: int = HTTP_401_UNAUTHORIZED,
        detail: str = "Unauthorized",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code, detail, headers)


class MFARequired(Unauthorized):
    """
    An exception raised to stop a request execution and return a 401 HTTP Error for MFA.
    """

    def __init__(
        self,
        status_code: int = HTTP_401_UNAUTHORIZED,
        detail: str = "MFA Required",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code, detail, headers)


class Forbidden(HTTPException):
    """An exception raised to stop a request execution and return a 403 HTTP Error."""

    def __init__(
        self,
        status_code: int = HTTP_403_FORBIDDEN,
        detail: str = "Forbidden",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code, detail, headers)


class BoolMeta(type):
    def __bool__(cls) -> bool:
        return cls.kind == "allowed"  # type: ignore

    def __repr__(cls) -> str:
        return cls.reason  # type: ignore


class HasPermission(int, metaclass=BoolMeta):
    """
    A type used to know if a permission is allowed or not.

    It behave has a boolean, but 3 possibilities exists defind has 3 sub-types
    {class}`Allowed` {class}`Unauthenticated` or {class}`Denied`.

    In many cases Unauthenticated call may redirect to a login page,
    where authenticated user are not redirected. they have an error message,
    or the frontend may use the information to adapt its interface.
    """

    kind: Literal["allowed", "unauthenticated", "mfa_required", "denied"]
    """
    Identified basic information of the response.
    It distinguished unauthenticated and denied to eventually raised 401 over 403 error.
    """
    reason: str
    """A human explanation of the response."""

    def __new__(cls, reason: str) -> "HasPermission":
        instance = super().__new__(cls)
        instance.reason = reason
        return instance

    def __repr__(self) -> str:
        return self.reason

    def __bool__(self) -> bool:
        return self.kind == "allowed"


class Allowed(HasPermission):
    """Represent a permission check result that is allowed."""

    kind = "allowed"
    reason = "Allowed"


class Unauthenticated(HasPermission):
    """
    Represent a permission check result that is not allowed due to
    missing authentication mechanism.
    """

    kind = "unauthenticated"
    reason = "Authentication required"


class PreAuthenticated(HasPermission):
    """
    Represent a permission check result that is not allowed due to
    missing secondary authentication mechanism.
    """

    kind = "mfa_required"
    reason = "MFA required"


class Denied(HasPermission):
    """
    Represent a permission check result that is not allowed due to lack of permission.
    """

    kind = "denied"
    reason = "Access denied to this resource"



"""HTML Form helpers."""

from collections.abc import Sequence
from inspect import isclass
from typing import Any, get_args, get_origin

from pydantic import BaseModel, ValidationError

from fastlife.shared_utils.infer import (
    get_runtime_type,
    get_type_by_discriminator,
    is_union,
)


def flatten_error(
    exc: ValidationError, prefix: str, pydantic_model: type[Any]
) -> dict[str, str]:
    """
    Flatten a Pydantic ValidationError into a dictionary of error messages.

    The dict returned use the same key as the way
    :meth:`pydantic_form <fastlife.adapters.xcomponent.renderer.XRendererFactory.pydantic_form>`
    serialized its data, in order to display error message on appropriate fields
    on the html representation.

    :param exc: The ValidationError exception containing validation errors
    :param prefix: The initial prefix/path for the error location
    :param pydantic_model: The Pydantic model class being validated
    :return: Dictionary mapping error locations to error messages
    """
    errors: dict[str, str] = {}
    typ: Any
    runtime_type: Any = get_runtime_type(pydantic_model)
    for error in exc.errors():
        loc = prefix
        typ = runtime_type
        locations = iter(error["loc"])
        while True:
            part = next(locations, None)
            if part is None:
                # safety break that should not happen
                break  # coverage: ignore
            if isinstance(part, str):
                assert isclass(typ) and issubclass(typ, BaseModel)  # type: ignore
                field = typ.model_fields[part]
                assert field.annotation
                typ = get_runtime_type(field.annotation)
                type_origin = get_origin(typ)
                if type_origin:
                    if is_union(typ):
                        assert isinstance(field.discriminator, str)
                        loc = f"{loc}.{part}"
                        part = next(locations, None)
                        if part is None:
                            break
                        typ = get_type_by_discriminator(part, field.discriminator, typ)
                        assert typ is not None  # pydantic never let you get None here
                    else:
                        loc = f"{loc}.{part}"
                elif issubclass(typ, BaseModel):
                    loc = f"{loc}.{part}"
                else:
                    # the last attribute
                    loc = f"{loc}.{part}"
                    break
            else:
                # we are in a sequence
                loc = f"{loc}.{part}"
                type_origin = get_origin(typ)
                assert type_origin and issubclass(type_origin, Sequence)
                typ = get_args(typ)[0]

        if loc in errors:
            errors[loc] = f"{errors[loc]}, {error['msg']}"
        else:
            errors[loc] = error["msg"]
    return errors



"""Handle Enum type."""

from collections.abc import Mapping
from enum import Enum
from typing import Any

from pydantic.fields import FieldInfo

from fastlife.adapters.xcomponent.pydantic_form.widget_factory.base import (
    BaseWidgetBuilder,
)
from fastlife.adapters.xcomponent.pydantic_form.widgets.base import Widget
from fastlife.adapters.xcomponent.pydantic_form.widgets.dropdown import DropDownWidget


class EnumBuilder(BaseWidgetBuilder[Enum]):
    """Builder for Enum."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for Enum."""
        return issubclass(typ, Enum)

    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],  # an enum subclass
        field: FieldInfo | None,
        value: Enum | None,  # str | int | float,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> Widget[Enum]:
        """Build the widget."""
        options = [
            (str(item.value), str(item.value))  # type: ignore
            for item in field_type  # type: ignore
        ]
        if isinstance(value, str):
            # if the form was invalid, we rebuild the enum with a model
            # that has not been validate, so we rebuild the type here.
            try:
                value = field_type(value)
            except ValueError:
                # we are neither supposed to failed nor responsible to report the error
                value = None
        return DropDownWidget(
            name=field_name,
            options=options,  # type: ignore
            removable=removable,
            title=field.title or "" if field else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            token=self.factory.token,
            value=str(value.value) if value else "",
            error=form_errors.get(field_name),
        )



"""Application registry."""

from collections.abc import AsyncIterator, Mapping
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from fastapi import FastAPI

if TYPE_CHECKING:
    from fastlife.service.locale_negociator import LocaleNegociator  # coverage: ignore
    from fastlife.service.request_factory import RequestFactory  # coverage: ignore
    from fastlife.service.templates import (  # coverage: ignore
        AbstractTemplateRendererFactory,  # coverage: ignore
    )  # coverage: ignore
    from fastlife.service.translations import LocalizerFactory  # coverage: ignore

from fastlife.domain.model.template import InlineTemplate
from fastlife.settings import Settings

TSettings = TypeVar("TSettings", bound=Settings, covariant=True)
"""
A TypeVar used to override the DefaultRegistry to add more helpers in the registry.
"""


class GenericRegistry(Generic[TSettings]):
    """
    Application registry for fastlife dependency injection.
    It is initialized by the configurator and accessed by the `fastlife.Registry`.
    """

    settings: TSettings
    """Application settings."""
    renderers: Mapping[type[InlineTemplate], "AbstractTemplateRendererFactory[Any]"]
    """Registered template engine."""
    locale_negociator: "LocaleNegociator"
    """Used to fine the best language for the response."""
    localizer: "LocalizerFactory"
    """Used to localized message."""
    request_factory: "RequestFactory"

    def __init__(self, settings: TSettings) -> None:
        from fastlife.service.job import JobScheduler
        from fastlife.service.locale_negociator import default_negociator
        from fastlife.service.request_factory import default_request_factory
        from fastlife.service.translations import LocalizerFactory

        self.settings = settings
        self.locale_negociator = default_negociator(self.settings)
        self.renderers = {}
        self.localizer = LocalizerFactory()
        self.request_factory = default_request_factory(self)
        self.job_scheduler = JobScheduler(self)

    def get_renderer(
        self, template: InlineTemplate
    ) -> "AbstractTemplateRendererFactory[Any]":
        for cls in template.__class__.__mro__:
            if cls in self.renderers:
                return self.renderers[cls]
        raise RuntimeError(
            f"No renderer registered for template {template.__class__.__qualname__}"
        )

    @asynccontextmanager
    async def lifespan(self, app: FastAPI) -> AsyncIterator[Any]:
        """
        The [lifespan](https://asgi.readthedocs.io/en/latest/specs/lifespan.html)
        is used to start the job scheduler.

        In fastlife the application state is the registry, it has to be overriden
        to add an implementation.
        """
        self.job_scheduler.start()
        yield
        # FIXME: make the wait statement comming from the settings
        self.job_scheduler.shutdown(wait=True)


DefaultRegistry = GenericRegistry[Settings]
"""
The default registry until you need to inject more component in the registry.
"""


TRegistry = TypeVar("TRegistry", bound=DefaultRegistry, covariant=True)
"""
A TypeVar used to override the DefaultRegistry to add more helpers in the registry.
"""



"""Table components."""

from collections.abc import Mapping

from xcomponent import XNode

from fastlife.adapters.xcomponent.registry import BUILTINS_CATALOG_NS, x_component


@x_component(namespace=BUILTINS_CATALOG_NS)
def Table(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    html ``<table>`` node.

    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.TABLE_CLASS`.
    """
    return """
    <table id={id} class={class_ or globals.TABLE_CLASS}>
        {children}
    </table>
    """


@x_component(namespace=BUILTINS_CATALOG_NS)
def Thead(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    html ``<thead>`` node.

    :param id: unique identifier of the element.
    :param class_: css class for the node
    """
    return """
    <thead id={id} class={class_}>
        {children}
    </thead>
    """


@x_component(namespace=BUILTINS_CATALOG_NS)
def Tbody(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    html ``<tbody>`` node.

    :param id: unique identifier of the element.
    :param class_: css class for the node
    """
    return """
    <tbody id={id} class={class_}>
        {children}
    </tbody>
    """


@x_component(namespace=BUILTINS_CATALOG_NS)
def Tfoot(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    html ``<tfoot>`` node.

    :param id: unique identifier of the element.
    :param class_: css class for the node
    """
    return """
    <tfoot id={id} class={class_}>
        {children}
    </tfoot>
    """


@x_component(namespace=BUILTINS_CATALOG_NS)
def Th(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    html ``<th>`` node.

    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.TH_CLASS`.
    """
    return """
    <th id={id} class={class_ or globals.TH_CLASS}>
        {children}
    </th>
    """


@x_component(namespace=BUILTINS_CATALOG_NS)
def Td(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    html ``<td>`` node.

    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.TD_CLASS`.
    """
    return """
    <td id={id} class={class_ or globals.TD_CLASS}>
        {children}
    </td>
    """



"""
CLI Application Integration
"""

import inspect
from asyncio import run
from collections.abc import Callable
from functools import wraps
from typing import Any, Generic

from typer import Typer
from typer.core import DEFAULT_MARKUP_MODE, MarkupMode, TyperCommand, TyperGroup
from typer.models import Default

from fastlife import TRegistry
from fastlife.adapters.typer.model import CLIHook


class AsyncTyper(Typer, Generic[TRegistry]):
    """
    An typer class made for fastlife application.

    Look at the Typer documentation for the details, this class inherits from it.
    """

    registry: TRegistry
    """Access to the application registry."""

    def __init__(
        self,
        registry: TRegistry,
        *,
        name: str | None = Default(None),
        cls: type[TyperGroup] | None = Default(None),  # noqa: B008
        invoke_without_command: bool = Default(False),
        no_args_is_help: bool = Default(False),
        subcommand_metavar: str | None = Default(None),
        chain: bool = Default(False),
        result_callback: Callable[..., Any] | None = Default(None),  # noqa: B008
        # Command
        context_settings: dict[Any, Any] | None = Default(None),  # noqa: B008
        callback: Callable[..., Any] | None = Default(None),  # noqa: B008
        help: str | None = Default(None),
        epilog: str | None = Default(None),
        short_help: str | None = Default(None),
        options_metavar: str = Default("[OPTIONS]"),
        add_help_option: bool = Default(True),
        hidden: bool = Default(False),
        deprecated: bool = Default(False),
        add_completion: bool = True,
        # Rich settings
        rich_markup_mode: MarkupMode = DEFAULT_MARKUP_MODE,
        rich_help_panel: str | None = Default(None),
        suggest_commands: bool = True,
        pretty_exceptions_enable: bool = True,
        pretty_exceptions_show_locals: bool = True,
        pretty_exceptions_short: bool = True,
    ):
        self.registry = registry
        super().__init__(
            name=name,
            cls=cls,
            invoke_without_command=invoke_without_command,
            no_args_is_help=no_args_is_help,
            subcommand_metavar=subcommand_metavar,
            chain=chain,
            result_callback=result_callback,
            context_settings=context_settings,
            callback=callback,
            help=help,
            epilog=epilog,
            short_help=short_help,
            options_metavar=options_metavar,
            add_help_option=add_help_option,
            hidden=hidden,
            deprecated=deprecated,
            add_completion=add_completion,
            rich_markup_mode=rich_markup_mode,
            rich_help_panel=rich_help_panel,
            suggest_commands=suggest_commands,
            pretty_exceptions_enable=pretty_exceptions_enable,
            pretty_exceptions_show_locals=pretty_exceptions_show_locals,
            pretty_exceptions_short=pretty_exceptions_short,
        )

    def command(
        self,
        name: str | None = None,
        *,
        cls: type[TyperCommand] | None = None,
        context_settings: dict[Any, Any] | None = None,
        help: str | None = None,
        epilog: str | None = None,
        short_help: str | None = None,
        options_metavar: str | None = None,
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
        rich_help_panel: str | None = Default(None),
    ) -> CLIHook:
        """Override the Typer command to accept async functions."""

        def decorator(cmd: CLIHook) -> CLIHook:
            # Get the original command signature
            sig = inspect.signature(cmd)
            # Remove the registry parameter from the signature
            has_registry = False
            new_params: list[inspect.Parameter] = []
            for param_name, param in sig.parameters.items():
                if param_name == "registry":
                    has_registry = True
                else:
                    new_params.append(param)

            def build_signature(
                original_func: Callable[..., Any], new_params: list[inspect.Parameter]
            ) -> inspect.Signature:
                """
                Create a new signature for the wrapped function for typer.

                We remove the registry if exists and then
                """
                return inspect.Signature(
                    parameters=new_params,
                    return_annotation=original_func.__annotations__.get("return"),
                )

            @wraps(
                cmd,
                assigned=(
                    "__module__",
                    "__name__",
                    "__qualname__",
                    "__doc__",
                    "__annotations__",
                ),
            )
            def run_command(**cli_kwargs: Any) -> Any:
                if has_registry:
                    cli_kwargs["registry"] = self.registry
                res = cmd(**cli_kwargs)
                if inspect.iscoroutine(res):
                    res = run(res)
                return res

            run_command.__signature__ = build_signature(  # type: ignore
                cmd, new_params
            )

            super(AsyncTyper, self).command(
                name=name,
                cls=cls,
                context_settings=context_settings,
                help=help,
                epilog=epilog,
                short_help=short_help,
                options_metavar=options_metavar,
                add_help_option=add_help_option,
                no_args_is_help=no_args_is_help,
                hidden=hidden,
                deprecated=deprecated,
                rich_help_panel=rich_help_panel,
            )(run_command)
            return cmd

        return decorator



"""Hidden fields"""

from fastlife.domain.model.types import Builtins

from .base import Widget


class HiddenWidget(Widget[Builtins]):
    '''
    Widget to annotate to display a field as an hidden field.

    ::
        from pydantic import BaseModel
        from fastlife.adapters.xcomponent.pydantic_form.widgets.base import CustomWidget
        from fastlife.adapters.xcomponent.pydantic_form.widgets.hidden import HiddenWidget

        class MyForm(BaseModel):
            id: Annotated[str, CustomWidget(HiddenWidget)] = Field(...)
            """Identifier in the database."""

    '''

    template = """
    <Hidden name={name} value={value} id={id} />
    """



"""
Views for pydantic form.

Pydantic form generate form that may contains fields that requires some ajax query.
"""

from fastapi import Query

from fastlife import Configurator, Request, Response, XTemplate, configure
from fastlife.domain.model.template import InlineTemplate
from fastlife.shared_utils.resolver import resolve_extended


async def show_widget(
    typ: str,
    request: Request,
    name: str | None = Query(None),
    token: str | None = Query(None),
    removable: bool = Query(False),
) -> Response:
    """
    This views is used by pydantic_form to generate a nested field asynchronously.
    """
    model_cls = resolve_extended(typ)

    template: InlineTemplate = XTemplate()

    rndr = request.registry.get_renderer(template)(request)
    lczr = request.registry.localizer(request.locale_name)
    rndr.globals.update({"request": request, **lczr.as_dict()})
    data = rndr.pydantic_form_field(
        model=model_cls,  # type: ignore
        name=name,
        token=token,
        removable=removable,
        field=None,
    )
    return Response(data, headers={"Content-Type": "text/html"})


@configure
def includeme(config: Configurator) -> None:
    route_prefix = config.registry.settings.fastlife_route_prefix
    config.add_route(
        "fl-pydantic-form-widget",
        f"{route_prefix}/pydantic-form/widgets/{{typ}}",
        show_widget,
        methods=["GET"],
    )



"""
Configure scheduled job using a decorator.

Scheduled jobs are useful for handling application lifecycle tasks,
such as database cleanup, cache population, and other asynchronous
operations that aren't tied to user requests.

Exemple of usage:

```python
from typing import Annotated

from fastapi import Response
from fastlife import scheduled_job


@scheduled_job(trigger="interval", seconds=60)
async def cleanup():
    with self.registry.uow_factory() as t:
        await t.uow.tokens.remove_expired_sessions()
        await t.commit()
```

Note that if the trigger parameter is comming from a settings,
the usage of a {func}`fastlife.config.configurator.configure` decorator
is more appropriate in order to inject it using the underlying method
{func}`fastlife.config.configurator.register_job` consumed by the
scheduled_job decorator.
"""

from collections.abc import Callable
from datetime import datetime
from typing import Any

import venusian
from apscheduler.util import undefined

from fastlife.service.job import JobHandler, JobSchedulerTrigger, Undefined
from fastlife.service.registry import TRegistry

from .configurator import VENUSIAN_CATEGORY, GenericConfigurator


def scheduled_job(
    *,
    trigger: JobSchedulerTrigger | None = None,
    id: str | None = None,
    name: str | None = None,
    misfire_grace_time: int | Undefined = undefined,
    coalesce: Undefined = undefined,
    max_instances: int | Undefined = undefined,
    next_run_time: datetime | Undefined = undefined,
    jobstore: str = "default",
    executor: str = "default",
    replace_existing: bool = False,
    **trigger_args: Any,
) -> Callable[..., JobHandler[TRegistry]]:
    """
    A decorator function to register a job in the
    {func}`Configurator <fastlife.config.configurator.GenericConfigurator.register_job>`
    while scaning a module using {func}`include
    <fastlife.config.configurator.GenericConfigurator.include>`.

    :param trigger: the way the job is triggered.
    :param id: optional identifier for the job.
    :param name: optional name for the job.
    :param misfire_grace_time: seconds after the designated runtime that the job is still
        allowed to be run (or `None` to allow the job to run no matter how late it is)
    :param coalesce: run once instead of many times if the scheduler determines that the
        job should be run more than once in succession
    :param max_instances: maximum number of concurrently running instances allowed for this
        job
    :param next_run_time: when to first run the job, regardless of the trigger (pass
        `None` to add the job as paused)
    :param jobstore: alias of the job store to store the job in
    :param executor: alias of the executor to run the job with
    :param replace_existing: `True` to replace an existing job with the same `id`
        (but retain the number of runs from the existing one)
    :return: the configuration callback.
    """
    job_name = name

    def configure(
        wrapped: JobHandler[TRegistry],
    ) -> JobHandler[TRegistry]:
        def callback(
            scanner: venusian.Scanner, name: str, ob: JobHandler[TRegistry]
        ) -> None:
            if not hasattr(scanner, VENUSIAN_CATEGORY):
                return  # coverage: ignore
            config: GenericConfigurator[TRegistry] = getattr(scanner, VENUSIAN_CATEGORY)
            config.register_job(
                wrapped,
                trigger=trigger,
                id=id,
                name=job_name or name,
                misfire_grace_time=misfire_grace_time,
                coalesce=coalesce,
                max_instances=max_instances,
                next_run_time=next_run_time,
                jobstore=jobstore,
                executor=executor,
                replace_existing=replace_existing,
                **trigger_args,
            )

        venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
        return wrapped

    return configure



from importlib import metadata

__version__ = metadata.version("fastlifeweb")

from .adapters.fastapi.form import form_model
from .adapters.fastapi.localizer import Localizer
from .adapters.fastapi.request import (
    AnyRequest,
    Registry,
    Request,
    get_registry,
    get_request,
)
from .adapters.fastapi.websocket import GenericWebSocket, WebSocket
from .adapters.xcomponent.pydantic_form.widgets.base import CustomWidget, Widget
from .adapters.xcomponent.pydantic_form.widgets.boolean import BooleanWidget
from .adapters.xcomponent.pydantic_form.widgets.checklist import ChecklistWidget
from .adapters.xcomponent.pydantic_form.widgets.dropdown import (
    DropDownWidget,
    DropDownWidgetOption,
)
from .adapters.xcomponent.pydantic_form.widgets.hidden import HiddenWidget
from .adapters.xcomponent.pydantic_form.widgets.mfa_code import MFACodeWidget
from .adapters.xcomponent.pydantic_form.widgets.model import ModelWidget
from .adapters.xcomponent.pydantic_form.widgets.sequence import SequenceWidget
from .adapters.xcomponent.pydantic_form.widgets.text import (
    PasswordWidget,
    TextareaWidget,
    TextWidget,
)
from .adapters.xcomponent.pydantic_form.widgets.union import UnionWidget
from .adapters.xcomponent.registry import x_component, x_function
from .config import (
    Configurator,
    GenericConfigurator,
    configure,
    exception_handler,
    resource,
    resource_view,
    scheduled_job,
    view_config,
    websocket_view,
)
from .config.cli_command import cli_command
from .domain.model.asgi import ASGIRequest, ASGIResponse
from .domain.model.form import FormModel
from .domain.model.request import GenericRequest
from .domain.model.response import RedirectResponse, Response
from .domain.model.security_policy import (
    Allowed,
    Anonymous,
    Authenticated,
    AuthenticationState,
    Denied,
    Forbidden,
    HasPermission,
    NoMFAAuthenticationState,
    PendingMFA,
    PreAuthenticated,
    TClaimedIdentity,
    TIdentity,
    Unauthenticated,
    Unauthorized,
)
from .domain.model.template import XTemplate

# from .request.form_data import model
from .service.registry import DefaultRegistry, GenericRegistry, TRegistry, TSettings
from .service.request_factory import RequestFactory
from .service.security_policy import (
    AbstractNoMFASecurityPolicy,
    AbstractSecurityPolicy,
    InsecurePolicy,
)
from .service.translations import TranslatableStringFactory
from .settings import Settings

__all__ = [
    # Config
    "GenericConfigurator",
    "GenericRegistry",
    "Registry",
    "Settings",
    "configure",
    "view_config",
    "websocket_view",
    "exception_handler",
    "resource",
    "resource_view",
    "Configurator",
    "DefaultRegistry",
    "TSettings",
    "TRegistry",
    "get_registry",
    # Form
    "FormModel",
    "form_model",
    # Request
    "GenericRequest",
    "AnyRequest",
    "Request",
    "get_request",
    # Request Factory
    "ASGIRequest",
    "ASGIResponse",
    "RequestFactory",
    # Response
    "Response",
    "RedirectResponse",
    # Websockets
    "GenericWebSocket",
    "WebSocket",
    # Security
    "AbstractSecurityPolicy",
    "AbstractNoMFASecurityPolicy",
    "HasPermission",
    "Unauthenticated",
    "PreAuthenticated",
    "Allowed",
    "Denied",
    "Unauthorized",
    "Forbidden",
    "InsecurePolicy",
    "Anonymous",
    "PendingMFA",
    "Authenticated",
    "AuthenticationState",
    "NoMFAAuthenticationState",
    "TClaimedIdentity",
    "TIdentity",
    # Template
    "XTemplate",
    "x_component",
    "x_function",
    # Widgets
    "CustomWidget",
    "Widget",
    "BooleanWidget",
    "ChecklistWidget",
    "DropDownWidget",
    "DropDownWidgetOption",
    "HiddenWidget",
    "MFACodeWidget",
    "ModelWidget",
    "SequenceWidget",
    "TextWidget",
    "PasswordWidget",
    "TextareaWidget",
    "UnionWidget",
    # i18n
    "Localizer",
    "TranslatableStringFactory",
    # scheduled jobs
    "scheduled_job",
    # cli command
    "cli_command",
]



"""Handle Union type."""

from collections.abc import Mapping
from types import NoneType
from typing import TYPE_CHECKING, Annotated, Any, get_args, get_origin

from pydantic import ValidationError, create_model
from pydantic.fields import FieldInfo

from fastlife.adapters.xcomponent.pydantic_form.widget_factory.base import (
    BaseWidgetBuilder,
)
from fastlife.adapters.xcomponent.pydantic_form.widgets.base import Widget
from fastlife.adapters.xcomponent.pydantic_form.widgets.union import UnionWidget
from fastlife.shared_utils.infer import is_complex_type, is_union

if TYPE_CHECKING:
    from fastlife.adapters.xcomponent.pydantic_form.widget_factory.factory import (
        WidgetFactory,
    )


def get_title(typ: type[Any]) -> str:
    title = typ.__name__
    if get_origin(typ) is Annotated:
        base, *meta = get_args(typ)
        title = base.__name__
        for arg in meta:
            if isinstance(arg, str):
                title = arg
                break
    return title


def get_title_from_discriminator(discriminant: str, unionfield: FieldInfo) -> str:
    # we assume we have unionfield.discriminator here,
    # and we may have not
    assert unionfield.annotation
    child_name = discriminant
    for child_typ in unionfield.annotation.__args__:
        title = get_title(child_typ)
        if get_origin(child_typ) is Annotated:
            child_typ = get_args(child_typ)[0]
        if (
            discriminant
            in child_typ.model_fields[unionfield.discriminator].annotation.__args__
        ):
            child_name = title
    return child_name


def get_type_from_discriminator(discriminator: str | int, unionfield: FieldInfo) -> Any:
    # we assume we have unionfield.discriminator here,
    # and we may have not
    assert unionfield.annotation
    for child_typ in unionfield.annotation.__args__:
        if get_origin(child_typ) is Annotated:
            child_typ = get_args(child_typ)[0]
        if (
            discriminator
            in child_typ.model_fields[unionfield.discriminator].annotation.__args__
        ):
            return child_typ
    return None


def get_child_widget(
    field: FieldInfo | None,
    value: Any,
    factory: "WidgetFactory",
    field_name: str,
    form_errors: Mapping[str, Any],
) -> Widget[Any] | None:
    child = None
    if value:
        discriminant: Any = None
        dicriminated_type: Any = None
        assert field and field.annotation and isinstance(field.discriminator, str)
        DynamicModel = create_model("DynamicModel", value=(field.annotation, field))
        try:
            submod = DynamicModel(value=value)
            discriminant = getattr(submod.value, field.discriminator)  # type: ignore
            dicriminated_type = type(submod.value)  # type: ignore
            value = submod.value.model_dump()  # type: ignore
        except ValidationError as exc:
            submod = DynamicModel.model_construct(value=value)
            loc = exc.errors()[0]["loc"]
            if len(loc) > 1:
                discriminant = loc[1]
                dicriminated_type = get_type_from_discriminator(discriminant, field)
            else:
                # the discriminator failed, is there
                # a better option that just drop the value ?
                value = None  # coverage: ignore

        if isinstance(discriminant, str) and value:
            child = factory.build(
                dicriminated_type,
                name=field_name,
                field=FieldInfo(
                    title=get_title_from_discriminator(discriminant, field)
                ),
                value=value,  # type: ignore
                form_errors=form_errors,
                removable=False,
            )
    return child


class UnionBuilder(BaseWidgetBuilder[Any]):
    """Builder for Union."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for unions Union[A,B], A | B or event Optional[A], A | None"""
        return is_union(typ)

    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],
        field: FieldInfo | None,
        value: Any | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> Widget[Any]:
        """Build the widget."""
        types: dict[type[Any], str] = {}
        # required = True
        for typ in field_type.__args__:  # type: ignore
            if typ is NoneType:
                # required = False
                continue

            title = get_title(typ)
            if get_origin(typ) is Annotated:
                typ = get_args(typ)[0]

            types[typ] = title

        if (
            not removable
            and len(types) == 1
            # if the optional type is a complex type,
            and not is_complex_type(next(iter(types.keys())))
        ):
            return self.factory.build(  # coverage: ignore
                next(iter(types.keys())),
                name=field_name,
                field=field,
                value=value,
                form_errors=form_errors,
                removable=False,
            )

        child = get_child_widget(
            field,
            value,
            self.factory,
            field_name,
            form_errors,
        )

        # FIXME Union[Sequence[FooModel]]
        # if isinstance(child, Sequence):
        #     child = child.__args__[0]
        widget = UnionWidget[Any](
            name=field_name,
            # we assume those types are BaseModel
            value=child,
            children_types={v: k for k, v in types.items()},
            title=field.title or "" if field else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            token=self.factory.token,
            removable=removable,
            error=form_errors.get(field_name),
        )

        return widget



"""
Configure views using a decorator.

A simple usage:

```python
from typing import Annotated

from fastapi import Response
from fastlife import Template, template, view_config


@view_config("hello_world", "/", methods=["GET"])
async def hello_world(
    template: Annotated[Template, template("HelloWorld.jinja")],
) -> Response:
    return template()
```
"""

from typing import Any

import venusian
from typer.core import TyperCommand
from typer.models import Default

from fastlife.adapters.typer.model import CLICommand, CLIHook

from .configurator import VENUSIAN_CATEGORY, Configurator


def cli_command(
    name: str | None = None,
    *,
    cls: type[TyperCommand] | None = None,
    context_settings: dict[Any, Any] | None = None,
    help: str | None = None,
    epilog: str | None = None,
    short_help: str | None = None,
    options_metavar: str | None = None,
    add_help_option: bool = True,
    no_args_is_help: bool = False,
    hidden: bool = False,
    deprecated: bool = False,
    rich_help_panel: str | None = Default(None),
) -> CLIHook:
    """
    A decorator function to register a websocket view in the
    {class}`Configurator <fastlife.config.configurator.GenericConfigurator>`
    while scaning a module using {func}`include
    <fastlife.config.configurator.GenericConfigurator.include>`.

    :return: the configuration callback.
    """
    command_name = name

    def configure(
        wrapped: CLIHook,
    ) -> CLIHook:
        def callback(scanner: venusian.Scanner, name: str, ob: CLIHook) -> None:
            if not hasattr(scanner, VENUSIAN_CATEGORY):
                return  # coverage: ignore
            config: Configurator = getattr(scanner, VENUSIAN_CATEGORY)
            config.add_cli_command(
                CLICommand(
                    hook=wrapped,
                    name=command_name,
                    cls=cls,
                    context_settings=context_settings,
                    help=help,
                    epilog=epilog,
                    short_help=short_help,
                    options_metavar=options_metavar,
                    add_help_option=add_help_option,
                    no_args_is_help=no_args_is_help,
                    hidden=hidden,
                    deprecated=deprecated,
                    rich_help_panel=rich_help_panel,
                )
            )

        venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
        return wrapped

    return configure



"""Customize the request class."""

from collections.abc import Callable
from typing import Any

from fastlife.domain.model.asgi import ASGIRequest
from fastlife.domain.model.request import GenericRequest
from fastlife.service.registry import DefaultRegistry, TRegistry

RequestFactory = Callable[[ASGIRequest], GenericRequest[Any, Any, Any]]
"""
Transform the [ASGIRequest](#fastlife.domain.model.asgi.ASGIRequest)
object to the fastlife [GenericRequest](#fastlife.domain.model.request.GenericRequest).
"""

RequestFactoryBuilder = Callable[[TRegistry], RequestFactory]
"""Interface to implement to create a request factory"""


def default_request_factory(registry: DefaultRegistry) -> RequestFactory:
    """The default request factory the return the generic request."""

    def request(request: ASGIRequest) -> GenericRequest[Any, Any, Any]:
        return GenericRequest[Any, Any, Any](registry, request)

    return request



from xcomponent import XNode

from fastlife.adapters.xcomponent.registry import PYDANTICFORM_CATALOG_NS, x_component


@x_component(namespace=PYDANTICFORM_CATALOG_NS)
def ErrorText(
    text: str,
    class_: str,
) -> str:
    """
    display an error for a field.

    :param text: error message
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.ERROR_CLASS`.
    """
    return """
        <span class={class_ or globals.ERROR_CLASS}>{text}</span>
    """


@x_component(namespace=PYDANTICFORM_CATALOG_NS)
def OptionalErrorText(
    text: str | None,
    class_: str | None = None,
) -> str:
    """
    display an error for a field.

    :param text: error message
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.ERROR_CLASS`.
    """
    return """
    <>
        {
            if text {
                <ErrorText text={text} class={class_} />
            }
        }
    </>
    """


@x_component(namespace=PYDANTICFORM_CATALOG_NS)
def FatalError(
    message: str | None,
    class_: str | None = None,
    icon_class: str | None = None,
    text_class: str | None = None,
) -> str:
    """
    display an error for a field.

    :param message: error message
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.FATAL_ERROR_CLASS`.
    :param icon_class: css class for the node, defaults to :attr:`fastlife.template_globals.Globals.FATAL_ERROR_ICON_CLASS`.
    :param text_class: css class for the node, defaults to :attr:`fastlife.template_globals.Globals.FATAL_ERROR_TEXT_CLASS`.
    """
    return """
    <>
    {
      if message {
        <div class={class_ or globals.FATAL_ERROR_CLASS} role="alert">
          <Icon name="fire" class={icon_class or globals.FATAL_ERROR_ICON_CLASS} />
          <span class={text_class or globals.FATAL_ERROR_TEXT_CLASS}>{message}</span>
        </div>
      }
    }
    </>
    """


@x_component(namespace=PYDANTICFORM_CATALOG_NS)
def Hint(text: str | None) -> str:
    """
    Display a hint message for a field.

    :param text: hint text.
    """
    return """
    <>
    {
        if text {
            <span class="mt-2 text-sm text-neutral-500 dark:text-neutral-400">
                {text}
            </span>
        }
    }
    </>
    """


@x_component(namespace=PYDANTICFORM_CATALOG_NS)
def Widget(
    widget_id: str,
    removable: bool,
    children: XNode,
) -> str:
    """
    Base component for widget.

    :param widget_id: widget to display.
    :param removable: Set to true to add a remove button
    """
    return """
    <div id={widget_id + "-container"}>
      {children}
      {
        if removable {
          <div class="my-4">
            <Button type="button"
                onclick={"document.getElementById('" + widget_id + "-container').remove()"}
                >
                Remove
            </Button>
         </div>
        }
      }
    </div>
    """



from datetime import date, datetime
from typing import Any

from pydantic import Field, field_validator

from .base import Widget


class DateWidget(Widget[date]):
    """
    Widget for text like field (email, ...).
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div class="pt-4">
        <Label for={id}>{globals.gettext(title)}</Label>
        <OptionalErrorText text={error} />
        <Date name={name} value={value} type={input_type} id={id}
          aria-label={aria_label} min={min} max={max} />
        <Hint text={hint} />
      </div>
    </Widget>
    """

    input_type: str = Field(default="date")
    """type attribute for the Input component."""
    min: date | None = Field(default=None)
    max: date | None = Field(default=None)


class DateTimeWidget(Widget[datetime]):
    """
    Widget for datetime.

    Important, is uses datetime-local from HTML5,
    which is not timezone aware.
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div class="pt-4">
        <Label for={id}>{globals.gettext(title)}</Label>
        <OptionalErrorText text={error} />
        <DateTime name={name} value={value} type={input_type} id={id}
          aria-label={aria_label} min={min} max={max} />
        <Hint text={hint} />
      </div>
    </Widget>
    """

    input_type: str = Field(default="datetime-local")
    """type attribute for the Input component."""
    min: date | None = Field(default=None)
    max: date | None = Field(default=None)

    @field_validator("value", mode="before")
    def validate_value(cls, value: Any) -> Any:
        if not value:
            return None
        if isinstance(value, str) and len(value) == 16:
            value += ":00"
        return value



"""
Configure views using a decorator.

A simple usage:

```python
from typing import Annotated

from fastapi import Response
from fastlife import Template, template, view_config


@view_config("hello_world", "/", methods=["GET"])
async def hello_world(
    template: Annotated[Template, template("HelloWorld.jinja")],
) -> Response:
    return template()
```
"""

from collections.abc import Callable
from typing import Any

import venusian

from .configurator import VENUSIAN_CATEGORY, Configurator


def websocket_view(
    name: str,
    path: str,
) -> Callable[..., Any]:
    """
    A decorator function to register a websocket view in the
    {class}`Configurator <fastlife.config.configurator.GenericConfigurator>`
    while scaning a module using {func}`include
    <fastlife.config.configurator.GenericConfigurator.include>`.

    Note that the current implementation does not support permission check using
    the security policy.

    :param name: name of the route, used to build route from the helper
        {meth}`fastlife.request.request.Request.url_for` in order to create links.
    :param path: path of the route, use `{curly_brace}` to inject FastAPI Path
        parameters.

    :return: the configuration callback.
    """
    view_name = name

    def configure(
        wrapped: Callable[..., Any],
    ) -> Callable[..., Any]:
        def callback(
            scanner: venusian.Scanner, name: str, ob: Callable[..., Any]
        ) -> None:
            if not hasattr(scanner, VENUSIAN_CATEGORY):
                return  # coverage: ignore
            config: Configurator = getattr(scanner, VENUSIAN_CATEGORY)
            config.add_websocket_route(
                name=view_name,
                path=path,
                endpoint=ob,
            )

        venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
        return wrapped

    return configure



"""Security policy permission routine."""

from collections.abc import Callable, Coroutine
from typing import Any

CheckPermissionHook = Callable[..., Coroutine[Any, Any, None]] | Callable[..., None]
CheckPermission = Callable[[str], CheckPermissionHook]


def check_permission(permission_name: str) -> CheckPermissionHook:
    """
    A closure that check that a user as the given permission_name.

    Adding a permission on the route requires that a security policy has been
    added using the method
    {meth}`fastlife.config.configurator.GenericConfigurator.set_security_policy`

    :param permission_name: a permission name set in a view to check access.
    :return: a function that raise http exceptions or any configured exception here.
    """
    # the check_permission is called by the configurator
    # and the request is exposed in the public module creating a circular dependency.
    from fastlife import Request  # a type must be resolved to inject a dependency.

    async def depencency_injection(request: Request) -> None:
        if request.security_policy is None:
            raise RuntimeError(
                f"Request {request.url.path} require a security policy, "
                "at least explicit fastlife.service.security_policy.InsecurePolicy "
                "is required"
            )
        allowed = await request.security_policy.has_permission(permission_name)
        match allowed.kind:
            case "allowed":
                return
            case "denied":
                raise request.security_policy.Forbidden(detail=allowed.reason)
            case "mfa_required":
                raise request.security_policy.MFARequired(detail=allowed.reason)
            case "unauthenticated":
                raise request.security_policy.Unauthorized(detail=allowed.reason)

    return depencency_injection



import logging

from fastlife.domain.model.asgi import ASGIApp, Message, Receive, Scope, Send
from fastlife.middlewares.base import AbstractMiddleware

log = logging.getLogger(__name__)


class XBackendTag(AbstractMiddleware):
    def __init__(
        self, app: ASGIApp, *, tag: str, header_name: str = "x-backend-tag"
    ) -> None:
        self.app = app
        self.header_name = header_name.encode()
        self.tag = tag.encode()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        async def send_wrapper(message: Message) -> None:
            if message["type"] == "http.response.start":
                headers = message.setdefault("headers", [])
                headers.append((self.header_name, self.tag))
            await send(message)

        await self.app(scope, receive, send_wrapper)



"""Handle Set type."""

from collections.abc import Mapping
from enum import Enum
from typing import Any, Literal, get_origin

from pydantic.fields import FieldInfo

from fastlife.adapters.xcomponent.pydantic_form.widget_factory.base import (
    BaseWidgetBuilder,
)
from fastlife.adapters.xcomponent.pydantic_form.widgets.base import Widget
from fastlife.adapters.xcomponent.pydantic_form.widgets.checklist import (
    Checkable,
    ChecklistWidget,
)


class SetBuilder(BaseWidgetBuilder[set[Any]]):
    """Builder for Set."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for Set"""
        return origin is set

    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],
        field: FieldInfo | None,
        value: set[Any] | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> Widget[Any]:
        """Build the widget."""
        choice_wrapper = field_type.__args__[0]
        choices = []
        choice_wrapper_origin = get_origin(choice_wrapper)
        if choice_wrapper_origin:
            if choice_wrapper_origin is Literal:
                litchoice: list[str] = choice_wrapper.__args__  # type: ignore
                choices = [
                    Checkable(
                        label=c,
                        value=c,
                        checked=c in value if value else False,  # type: ignore
                        name=field_name,
                        token=self.factory.token,
                        error=form_errors.get(f"{field_name}-{c}"),
                    )
                    for c in litchoice
                ]

            else:
                raise NotImplementedError  # coverage: ignore
        elif issubclass(choice_wrapper, Enum):
            choices = [
                Checkable(
                    label=e.value,
                    value=e.value,
                    checked=e in value if value else False,  # type: ignore
                    name=field_name,
                    token=self.factory.token,
                    error=form_errors.get(f"{field_name}-{e.name}"),
                )
                for e in choice_wrapper
            ]
        else:
            raise NotImplementedError  # coverage: ignore

        return ChecklistWidget(
            name=field_name,
            title=field.title or "" if field else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            token=self.factory.token,
            value=choices,
            removable=removable,
            error=form_errors.get(field_name),
        )



"""Type inference."""

import inspect
from collections.abc import Callable
from types import UnionType
from typing import Annotated, Any, Union, get_args, get_origin

from pydantic import BaseModel

from fastlife.domain.model.template import InlineTemplate


def is_newtype(typ: type[Any]) -> bool:
    """Check tha a type is a typing.NewType"""
    return callable(typ) and hasattr(typ, "__supertype__")


def get_runtime_type(typ: type[Any]) -> type[Any]:
    """
    Unwrap the NewType if exists and return the wrapped type in that case.
    """
    if is_newtype(typ):
        typ = typ.__supertype__
    return typ


def get_type_by_discriminator(
    discriminant: str | int, discriminator: str, typ: type[Any]
) -> type[Any] | None:
    """
    Find a type from a union using a discriminator.

    :param discriminant: the value searched in the discriminator attribute.
    :param discriminator: the type attribute name, must be a literal.
    :param typ: a union type
    :return: The type of the union or None
    """
    for child_typ in get_args(typ):
        if discriminant in child_typ.model_fields[discriminator].annotation.__args__:
            if get_origin(child_typ) is Annotated:
                child_typ = get_args(child_typ)[0]
            return child_typ
    return None


def is_complex_type(typ: type[Any]) -> bool:
    """
    Used to detect complex type such as Mapping, Sequence and pydantic BaseModel.

    This method cannot be used outside pydantic serialization.
    """
    typ = get_runtime_type(typ)
    return bool(get_origin(typ) or issubclass(typ, BaseModel))


def is_union(typ: type[Any]) -> bool:
    """Used to detect unions like Optional[T], Union[T, U] or T | U."""
    type_origin = get_origin(get_runtime_type(typ))
    if type_origin:
        if type_origin is Union:  # Optional[T]
            return True

        if type_origin is UnionType:  # T | U
            return True
    return False


def is_inline_template_returned(endpoint: Callable[..., Any]) -> bool:
    """Test if a view, the endpoint return a template."""
    signature = inspect.signature(endpoint)
    return_annotation = signature.return_annotation

    if isinstance(return_annotation, type) and issubclass(
        return_annotation, InlineTemplate
    ):
        return True

    if is_union(return_annotation):
        return any(
            isinstance(arg, type) and issubclass(arg, InlineTemplate)
            for arg in get_args(return_annotation)
        )

    return False



"""
Template Constants injects as global variables in templates.

Constants are configurable using the setting :attrs:`xcomponent_global_catalog_class`,
in order to customize templates.

Those constants are heavy used to inject CSS classes in primary html element
that are bound to Jinja component.

"""

from pydantic import BaseModel


def space_join(*segments: str) -> str:
    return " ".join(segments)


class Globals(BaseModel):
    """Templates constants."""

    A_CLASS: str = space_join(
        "text-primary-600",
        "hover:text-primary-500",
        "hover:underline",
        "dark:text-primary-300",
        "dark:hover:text-primary-400",
    )
    """Default css class for `A`."""

    BUTTON_CLASS: str = space_join(
        "appearance-none",
        "bg-primary-600",
        "px-5",
        "py-2.5",
        "font-semibold",
        "rounded-lg",
        "text-center",
        "text-sm",
        "text-white",
        "hover:bg-primary-700",
        "hover:bg-primary-200",
        "focus:outline-hidden",
        "focus:ring-4",
        "focus:ring-primary-300",
        "dark:bg-primary-600",
        "dark:focus:ring-primary-800",
        "dark:hover:bg-primary-700",
    )
    """Default css class for `Button`."""

    DETAILS_CLASS: str = "border border-neutral-100 p-4 rounded-m"
    """Default css class for `Details`."""

    SECONDARY_BUTTON_CLASS: str = space_join(
        "appearance-none",
        "bg-neutral-300",
        "px-5",
        "py-2.5",
        "rounded-lg",
        "font-semibold",
        "text-center",
        "text-neutral-900",
        "text-sm",
        "focus:outline-hidden",
        "focus:ring-4",
        "focus:ring-primary-300",
        "hover:bg-neutral-200",
        "dark:bg-neutral-400",
        "dark:focus:ring-neutral-100",
        "dark:hover:bg-neutral-300",
    )
    """
    css class for `Button`.

    usage:

    ```html
    <Button :class="SECONDARY_BUTTON_CLASS">secondary</Button>
    ```
    """

    ICON_BUTTON_CLASS: str = space_join(
        "appearance-none",
        "bg-white",
        "p-1",
        "rounded-xs",
        "text-primary-600",
        "border",
        "border-primary-600",
        "hover:bg-primary-200",
        "focus:outline-hidden",
        "focus:ring-4",
        "focus:ring-primary-500",
        "dark:bg-primary-900",
        "dark:text-neutral-300",
        "dark:hover:bg-primary-800",
        "dark:focus:ring-primary-500",
        "dark:hover:text-primary-300",
    )
    """
    css class for `Button`.

    usage:

    ```html
    <Button :class="ICON_BUTTON_CLASS">
      <icons.PencilSquare class="w-6 h-6" title="Copy" />
    </Button>

    ```
    """

    CHECKBOX_CLASS: str = space_join(
        "bg-neutral-100",
        "border-neutral-300",
        "h-4",
        "rounded-sm",
        "text-primary-600",
        "w-4",
        "dark:bg-neutral-700",
        "dark:border-neutral-600",
        "dark:focus:ring-primary-600",
        "dark:ring-offset-neutral-800",
        "focus:ring-2",
        "focus:ring-primary-500",
    )
    """Default css class for `Checkbox`."""

    FORM_CLASS: str = "space-y-4 md:space-y-6"
    """Default css class for `Form`."""

    H1_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-5xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-4xl",
    )
    """Default css class for `H1`."""

    H2_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-4xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-4xl",
    )
    """Default css class for `H2`."""

    H3_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-3xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-3xl",
    )
    """Default css class for `H3`."""

    H3_SUMMARY_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "text-3xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-3xl",
    )
    """
    Default css class for `H3` inside `Summary`.
    """

    H4_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-2xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-2xl",
    )
    """Default css class for `H4`."""

    H5_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-xl",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-xl",
    )
    """Default css class for `H5`."""

    H6_CLASS: str = space_join(
        "block",
        "font-bold",
        "font-sans",
        "leading-tight",
        "pb-4",
        "text-l",
        "text-neutral-900",
        "tracking-tight",
        "dark:text-white",
        "md:text-l",
    )
    """Default css class for `H6`."""

    INPUT_CLASS: str = space_join(
        "bg-neutral-50",
        "block",
        "border",
        "border-neutral-300",
        "p-2.5",
        "rounded-lg",
        "text-base",
        "text-neutral-900",
        "w-full",
        "dark:bg-neutral-700",
        "dark:border-neutral-600",
        "dark:focus:border-primary-500",
        "dark:focus:ring-primary-500",
        "dark:placeholder-neutral-400",
        "dark:text-white",
        "focus:border-primary-500",
        "focus:ring-primary-500",
    )
    """Default css class for `Input`."""

    LABEL_CLASS: str = space_join(
        "block",
        "font-bold",
        "mb-2",
        "text-base",
        "text-neutral-900",
        "dark:text-white",
    )
    """Default css class for `Label`."""

    P_CLASS: str = space_join(
        "text-base",
        "text-neutral-900",
        "dark:text-white",
    )
    """Default css class for `P`."""

    RADIO_DIV_CLASS: str = "flex items-center mb-4"
    """Default css class for `Radio` `<div>` container."""

    RADIO_INPUT_CLASS: str = space_join(
        "bg-neutral-100",
        "border-neutral-300",
        "w-4",
        "h-4",
        "text-primary-600",
        "focus:ring-2",
        "focus:ring-primary-500",
        "dark:bg-neutral-700",
        "dark:border-neutral-600",
        "dark:focus:ring-primary-600",
        "dark:ring-offset-neutral-800",
    )
    """
    Default css class for `Radio` `<input type="radio">`.
    """

    RADIO_LABEL_CLASS: str = (
        "ms-2 text-sm font-medium text-neutral-900 dark:text-neutral-300"
    )
    """
    Default css class for `Radio` `<label>` element.
    """

    SELECT_CLASS: str = space_join(
        "bg-neutral-50",
        "block",
        "border",
        "border-neutral-300",
        "p-2.5",
        "rounded-lg",
        "text-base",
        "text-neutral-900",
        "w-full",
        "focus:border-primary-500",
        "focus:ring-primary-500",
        "dark:bg-neutral-700",
        "dark:border-neutral-600",
        "dark:focus:border-primary-500",
        "dark:focus:ring-primary-500",
        "dark:placeholder-neutral-400",
        "dark:text-white",
    )
    """Default css class for `Select`."""

    SUMMARY_CLASS: str = "flex items-center items-center font-medium cursor-pointer"
    """Default css class for `Summary`."""

    TABLE_CLASS: str = "table-auto w-full text-left border-collapse"
    """Default css class for `Table`."""

    TD_CLASS: str = "px-4 py-2 font-normal border-b dark:border-neutral-500"
    """Default css class for `Td`."""

    TH_CLASS: str = "px-4 py-2 font-medium border-b dark:border-neutral-500"
    """Default css class for `Th`."""

    ERROR_CLASS: str = "mt-2 text-sm text-danger-500 dark:text-danger-400"
    """Default css class for `pydantic_form.Error`."""

    FATAL_ERROR_CLASS: str = (
        "flex items-center bg-red-50 border border-red-400 text-red-700"
    )
    """Default css class for `pydantic_form.FatalError`."""

    FATAL_ERROR_ICON_CLASS: str = "m-3 w-16 h-16 fill-orange-500"
    """Default css class for `pydantic_form.FatalError` icon."""

    FATAL_ERROR_TEXT_CLASS: str = "sm:inline text-xl"
    """Default css class for `pydantic_form.FatalError` text."""

    SUMMARY_OPEN_ICON_CLASS: str = (
        "w-8 h-8 transform transition-transform duration-300 rotate-90"
    )
    """Default css class for the svg icon of the open details/summary."""
    SUMMARY_CLOSED_ICON_CLASS: str = (
        "w-8 h-8 transform transition-transform duration-300"
    )
    """
    Default css class for the svg icon of the closed (collapsed) details/summary.
    """



"""
A middleware that update the request scope for https behind a proxy.

The attempt of this middleware is to fix Starlette behavior that use client and scheme
header based on the header ``x-forwarded-*`` headers and the ``x-real-ip``.

the ``x-forwarded-for`` header is not parsed to find the appropriate value,
the ``x-real-ip`` is used.
notethat the ``x-forwarded-port`` header is not used.

Note that uvicorn or hypercorn offer the same kind middleware.

Norw, every website is in https, so, this middleware is active by default.
"""

from fastlife import Configurator, configure

from .x_forwarded import XForwardedStar

__all__ = ["XForwardedStar"]


@configure
def includeme(config: Configurator) -> None:
    settings = config.registry.settings
    if settings.decode_reverse_proxy_headers:
        config.add_middleware(XForwardedStar)



"""Settings for the fastlife."""

from datetime import timedelta
from typing import Literal

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """
    Settings based on pydantic settings to configure the app.

    It can be overriden in order to inject application own settings.
    """

    model_config = SettingsConfigDict(env_prefix="fastlife_")
    """
    Set the prefix ``fastlife_`` for configuration using operating system environment.
    """

    fastlife_route_prefix: str = Field(default="/_fl")
    """Route prefix used for fastlife internal views."""

    registry_class: str = Field(default="fastlife.service.registry:DefaultRegistry")
    """Implementation class for the application regitry."""

    form_data_model_prefix: str = Field(default="payload")
    """
    Pydantic form default model prefix for serialized field in www-urlencoded-form.
    """
    csrf_token_name: str = Field(default="csrf_token")
    """
    Name of the html input field and for the http cookie for csrf token.
    """

    default_locale: str = Field(default="en")
    """
    The default locale
    """

    xcomponent_global_catalog_class: str = Field(
        default="fastlife.template_globals:Globals"
    )
    """
    Set global constants accessible in every templates.
    Defaults to `fastlife.template_globals:Globals`
    See {class}`fastlife.template_globals.Globals`
    """

    session_secret_key: str = Field(default="")
    """
    A secret key, that could not changes for life.

    Used to securize the session with itsdangerous.
    """
    session_cookie_name: str = Field(default="flsess")
    """Cookie name for the session cookie."""
    session_cookie_domain: str = Field(default="")
    """Cookie domain for the session cookie."""
    session_cookie_path: str = Field(default="/")
    """Cookie path for the session cookie."""
    session_duration: timedelta = Field(default=timedelta(days=14))
    """Cookie duration for the session cookie."""
    session_cookie_same_site: Literal["lax", "strict", "none"] = Field(default="lax")
    """Cookie same-site for the session cookie."""
    session_cookie_secure: bool = Field(default=False)
    """
    Cookie secure for the session cookie,

    should be true while using https on production.
    """
    session_serializer: str = Field(
        default="fastlife.adapters.itsdangerous:SignedSessionSerializer"
    )
    """Cookie serializer for the session cookie."""

    domain_name: str = Field(default="", title="domain name where the app is served")
    """Domain name whre the app is served."""

    decode_reverse_proxy_headers: bool = Field(default=True)
    """Ensure that the request object has information based on http proxy headers."""

    backend_tag_header_value: str = Field(
        default="", title="configure the XBackendTag middleware."
    )
    """
    Configure a value per instance to get the which machine has been reached
    by a load balancer.
    """

    backend_tag_header_name: str = Field(
        default="x-backend-tag", title="configure the XBackendTag middleware."
    )
    """
    HTTP Header for the {class}`fastlife.middlewares.backend_tag.XBackendTag`
    """

    scheduler_class: str = Field(
        default="apscheduler.schedulers.asyncio:AsyncIOScheduler"
    )
    """Implementation class of the job scheduler."""



"""
Adapt the XComponent template engine for fastlife.

It provide a list of html component ready to used and customizable,
and a few functions.

The pydantic_form bring form fieldset generation from pydantic base models.
"""
